# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
RiskIQ Threat Intelligence Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from datetime import datetime
from functools import partial
from typing import Any, Tuple, Optional, Union


from passivetotal import analyzer as ptanalyzer

from .ti_provider_base import (
    TIProvider,
    LookupResult,
    TISeverity,
    TILookupStatus,
    TIPivotProvider,
)
from ...common.utility import export
from ...common.exceptions import MsticpyUserError
from ..._version import VERSION

__version__ = VERSION
__author__ = "Mark Kendrick"


@export
class RiskIQ(TIProvider, TIPivotProvider):
    """RiskIQ Threat Intelligence Lookup."""

    _IOC_QUERIES: dict = {
        "ipv4": "ALL",
        "ipv4-articles": "articles",
        "ipv4-artifacts": "artifacts",
        "ipv4-certificates": "certificates",
        "ipv4-components": "components",
        "ipv4-cookies": "cookies",
        "ipv4-hostpairchildren": "hostpair_children",
        "ipv4-hostpairparents": "hostpair_parents",
        "ipv4-passivedns": "resolutions",
        "ipv4-projects": "projects",
        "ipv4-malware": "malware",
        "ipv4-rep": "reputation",
        "ipv4-services": "services",
        "ipv4-summary": "summary",
        "ipv4-trackers": "trackers",
        "ipv4-whois": "whois",
        "hostname": "ALL",
        "hostname-articles": "articles",
        "hostname-artifacts": "artifacts",
        "hostname-certificates": "certificates",
        "hostname-components": "components",
        "hostname-cookies": "cookies",
        "hostname-hostpairchildren": "hostpair_children",
        "hostname-hostpairparents": "hostpair_parents",
        "hostname-passivedns": "resolutions",
        "hostname-projects": "projects",
        "hostname-malware": "malware",
        "hostname-rep": "reputation",
        "hostname-summary": "summary",
        "hostname-trackers": "trackers",
        "hostname-whois": "whois",
    }

    # Aliases
    _IOC_QUERIES["dns"] = _IOC_QUERIES["hostname"]
    _IOC_QUERIES["dns-articles"] = _IOC_QUERIES["hostname-articles"]
    _IOC_QUERIES["dns-artifacts"] = _IOC_QUERIES["hostname-artifacts"]
    _IOC_QUERIES["dns-certificates"] = _IOC_QUERIES["hostname-certificates"]
    _IOC_QUERIES["dns-components"] = _IOC_QUERIES["hostname-components"]
    _IOC_QUERIES["dns-cookies"] = _IOC_QUERIES["hostname-cookies"]
    _IOC_QUERIES["dns-hostpairchildren"] = _IOC_QUERIES["hostname-hostpairchildren"]
    _IOC_QUERIES["dns-hostpairparents"] = _IOC_QUERIES["hostname-hostpairparents"]
    _IOC_QUERIES["dns-passivedns"] = _IOC_QUERIES["hostname-passivedns"]
    _IOC_QUERIES["dns-projects"] = _IOC_QUERIES["hostname-projects"]
    _IOC_QUERIES["dns-malware"] = _IOC_QUERIES["hostname-malware"]
    _IOC_QUERIES["dns-rep"] = _IOC_QUERIES["hostname-rep"]
    _IOC_QUERIES["dns-summary"] = _IOC_QUERIES["hostname-summary"]
    _IOC_QUERIES["dns-trackers"] = _IOC_QUERIES["hostname-trackers"]
    _IOC_QUERIES["dns-whois"] = _IOC_QUERIES["hostname-whois"]

    _PIVOT_ENTITIES = {
        prop: {"Dns": "DomainName", "IpAddress": "Address", "Host": "fqdn"}
        for prop in [
            "articles",
            "artifacts",
            "certificates",
            "components",
            "cookies",
            "hostpair_children",
            "hostpair_parents",
            "resolutions",
            "projects",
            "malware",
            "reputation",
            "summary",
            "trackers",
            "whois",
        ]
    }
    _PIVOT_ENTITIES["services"] = {"IpAddress": "Address"}

    _REFERENCE = "https://community.riskiq.com"

    def __init__(self, **kwargs):
        """Instantiate RiskIQ class."""
        super().__init__(**kwargs)
        ptanalyzer.init(username=kwargs.get("ApiID"), api_key=kwargs.get("AuthKey"))
        self._pivot_timespan_start: Optional[datetime] = None
        self._pivot_timespan_end: Optional[datetime] = None
        self._pivot_get_timespan: Any = None

    # pylint: disable=no-self-use
    @property
    def _requests_session(self):
        """Return the PT Analyzer session."""
        return ptanalyzer.api_clients["Cards"].session

    @_requests_session.setter
    def _requests_session(self, session):
        """Set the PT Analyzer session."""
        # pylint: disable=consider-using-dict-items
        for name in ptanalyzer.api_clients:
            ptanalyzer.api_clients[name].session = session

    # pylint: enable=no-self-use

    @staticmethod
    def _severity_rep(classification):
        """Get the severity level for a reputation score classification."""
        return {
            "MALICIOUS": TISeverity.high,
            "SUSPICIOUS": TISeverity.warning,
            "UNKNOWN": TISeverity.information,
            "GOOD": TISeverity.information,
        }.get(classification, TISeverity.information)

    def lookup_ioc(
        self, ioc: str, ioc_type: str = None, query_type: str = None, **kwargs
    ) -> LookupResult:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC Observable value
        ioc_type : str, optional
            IoC Type, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        LookupResult
            The returned results.

        """
        result = self._check_ioc_type(
            ioc=ioc, ioc_type=ioc_type, query_subtype=query_type
        )

        if result.status:
            return result

        result.provider = kwargs.get("provider_name", self.__class__.__name__)
        result.reference = self._REFERENCE

        if query_type is None:
            prop = "ALL"
        elif query_type not in [
            q.split("-", maxsplit=1)[-1] for q in self._IOC_QUERIES
        ]:
            result.result = False
            result.status = TILookupStatus.query_failed.value
            result.details = f"ERROR: unsupported query type {query_type}"
            return result
        else:
            prop = self._IOC_QUERIES.get(f"{result.ioc_type}-{query_type}", "ALL")

        try:
            ptanalyzer.set_context("msticpy", "ti", VERSION, prop)
            pt_obj = ptanalyzer.get_object(ioc)
            if prop == "ALL":
                result = self._parse_result_all_props(pt_obj, result)
            else:
                result = self._parse_result_prop(pt_obj, prop, result)
        except ptanalyzer.AnalyzerError as err:
            result.result = False
            result.status = TILookupStatus.query_failed.value
            result.details = f"ERROR: {err}"
            result.raw_result = err
            result.set_severity(TISeverity.unknown)

        return result

    def _parse_result_all_props(self, pt_result, ti_result):
        """Parse results for ALL properties."""
        ti_result.details = {
            "summary": pt_result.summary.as_dict,
            "reputation": pt_result.reputation.as_dict,
        }
        ti_result.raw_result = ti_result.details
        if pt_result.summary.total == 0 and pt_result.reputation.score == 0:
            ti_result.result = False
        else:
            ti_result.result = True
        rep_severity = self._severity_rep(pt_result.reputation.classification)
        ti_result.set_severity(rep_severity)
        if (
            "malware_hashes" in pt_result.summary.available
            or "articles" in pt_result.summary.available
        ):
            ti_result.set_severity(max(rep_severity, TISeverity.high))
        elif "projects" in pt_result.summary.available:
            ti_result.set_severity(max(rep_severity, TISeverity.warning))
        return ti_result

    def _parse_result_prop(self, pt_result, pt_prop, ti_result):
        """Parse result for a specific property."""
        attr = getattr(pt_result, pt_prop)
        if pt_prop == "reputation":
            ti_result.set_severity(self._severity_rep(attr.classification))
        elif pt_prop == "malware_hashes" and len(attr) > 0:
            ti_result.set_severity(TISeverity.high)
        else:
            ti_result.set_severity(TISeverity.information)
        ti_result.details = ti_result.raw_result = attr.as_dict
        ti_result.result = True
        return ti_result

    def parse_results(self, response: LookupResult) -> Tuple[bool, TISeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        Tuple[bool, TISeverity, Any]
            bool = positive or negative hit
            TISeverity = enumeration of severity
            Object with match details

        """
        return (True, TISeverity.information, None)

    def _set_pivot_timespan(self, **kwargs):
        """
        Set the pivot timespan and track whether it has changed.

        Returns
        -------
        bool
            whether the timespan changed.

        """
        changed = False
        start = kwargs.pop(
            "start",
            self._pivot_get_timespan().start if self._pivot_get_timespan else None,
        )
        end = kwargs.pop(
            "end", self._pivot_get_timespan().end if self._pivot_get_timespan else None
        )
        if (
            start
            and end
            and (start != self._pivot_timespan_start or end != self._pivot_timespan_end)
        ):
            changed = True
            self._pivot_timespan_start = start
            self._pivot_timespan_end = end
            ptanalyzer.set_date_range(start_date=start, end_date=end)
        return changed

    def pivot_value(self, prop, host, **kwargs):
        """Perform a pivot on a single value."""
        ts_changed = self._set_pivot_timespan(**kwargs)
        ptanalyzer.set_context("msticpy", "pivot", VERSION, prop)
        obj = ptanalyzer.get_object(host)
        if ts_changed and prop not in ["reputation", "summary", "whois"]:
            obj.reset(prop)
        try:
            attrib = getattr(obj, prop)
        except ptanalyzer.AnalyzerAPIError as err:
            raise RiskIQAPIUserError(err.message) from err
        except ptanalyzer.AnalyzerError as err:
            raise RiskIQUserError("Analyzer error.") from err
        return attrib.to_dataframe(**kwargs)

    def register_pivots(
        self,
        pivot_reg: "PivotRegistration",  # type: ignore # noqa: F821
        pivot: "Pivot",  # type: ignore # noqa: F821
    ):
        """
        Register pivot functions for the TI Provider.

        Parameters
        ----------
        pivot_reg : PivotRegistration
            Pivot registration settings.
        pivot : Pivot
            Pivot library instance

        """
        self._pivot_get_timespan = pivot.get_timespan
        self._pivot_timespan_start = None
        self._pivot_timespan_end = None
        base_reg = {
            "entity_container_name": "RiskIQ",
            "func_df_param_name": "data",
            "func_df_col_param_name": "host",
            "func_input_value_arg": "host",
            "func_out_column_name": "query",
        }
        for prop, entity_map in self._PIVOT_ENTITIES.items():
            reg = pivot_reg(
                func_new_name=prop,
                func_static_params={"prop": prop},
                input_type="value",
                entity_map=entity_map,
                **base_reg,
            )
            fun = partial(self.pivot_value)
            fun.__doc__ = getattr(
                ptanalyzer.Hostname, prop, getattr(ptanalyzer.IPAddress, prop)
            ).__doc__
            pivot.add_pivot_function(fun, pivot_reg=reg, container="RiskIQ")


class RiskIQUserError(MsticpyUserError):
    """Generic RiskIQ provider exception."""

    def __init__(
        self, *args, help_uri: Union[Tuple[str, str], str, None] = None, **kwargs
    ):
        """
        Create RiskIQ provider exception.

        Parameters
        ----------
        help_uri : Union[Tuple[str, str], str, None], optional
            Override the default help URI.

        """
        kwargs.update(title="error using RiskIQ python library")
        kwargs.update(
            ptlib_uri=(
                "RiskIQ PassiveTotal Python Library",
                "https://passivetotal.readthedocs.io",
            )
        )
        kwargs.update(
            riqinfo_uri=("RiskIQ Support", "https://www.riskiq.com/resources/support/")
        )
        uri = help_uri or self.DEF_HELP_URI
        super().__init__(*args, help_uri=uri, **kwargs)


class RiskIQAPIUserError(RiskIQUserError):
    """RiskIQ API provider exception."""

    def __init__(self, api_exception: ptanalyzer.AnalyzerAPIError):
        """
        Create RiskIQ API exception.

        Parameters
        ----------
        api_exception : ptanalyzer.AnalyzerAPIError
            Underlying API exception.

        """
        title = f"{api_exception.status_code} {api_exception.message}"
        super().__init__(title, str(api_exception))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mailbox Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class Mailbox(Entity):
    """
    Mailbox Entity class.

    Attributes
    ----------
    MailboxPrimaryAddress : str
        PrimaryAddress of the Mailbox
    DisplayName : str
        DisplayName of the Mailbox
    Upn : str
        Upn of the Mailbox
    ExternalDirectoryObjectId : str
        ExternalDirectoryObjectId of the Mailbox
    RiskLevel : str
        RiskLevel of the Mailbox

    """

    ID_PROPERTIES = ["MailboxPrimaryAddress"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.MailboxPrimaryAddress: Optional[str] = None
        self.DisplayName: Optional[str] = None
        self.Upn: Optional[str] = None
        self.ExternalDirectoryObjectId: Optional[str] = None
        self.RiskLevel: Optional[str] = None

        super().__init__(src_entity=src_entity, **kwargs)
        if src_event:
            self._create_from_event(src_event)

    def _create_from_event(self, src_event):
        self.MailboxPrimaryAddress = src_event.get("MailboxPrimaryAddress")
        self.Upn = src_event.get("Upn")
        self.DisplayName = src_event.get("DisplayName")

    @property
    def description_str(self):
        """Return Entity Description."""
        return (
            f"{self.MailboxPrimaryAddress} - {self.RiskLevel}"
            or self.__class__.__name__
        )

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.MailboxPrimaryAddress or self.__class__.__name__

    _entity_schema = {
        "MailboxPrimaryAddress": None,
        "DisplayName": None,
        "Upn": None,
        "ExternalDirectoryObjectId": None,
        "RiskLevel": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
AlienVault OTX Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from typing import Any, Tuple

import attr

from .ti_provider_base import LookupResult, TISeverity
from .http_base import HttpProvider, IoCLookupParams
from ...common.utility import export
from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods
@attr.s
class _OTXParams(IoCLookupParams):
    # override IoCLookupParams to set common defaults
    def __attrs_post_init__(self):
        self.headers = {"X-OTX-API-KEY": "{API_KEY}"}


@export
class OTX(HttpProvider):
    """AlientVault OTX Lookup."""

    _BASE_URL = "https://otx.alienvault.com"

    _IOC_QUERIES = {
        "ipv4": _OTXParams(path="/api/v1/indicators/IPv4/{observable}/general"),
        "ipv6": _OTXParams(path="/api/v1/indicators/IPv6/{observable}/general"),
        "ipv4-passivedns": _OTXParams(
            path="/api/v1/indicators/IPv4/{observable}/passive_dns"
        ),
        "ipv6-passivedns": _OTXParams(
            path="/api/v1/indicators/IPv6/{observable}/passive_dns"
        ),
        "ipv4-geo": _OTXParams(path="/api/v1/indicators/IPv4/{observable}/geo"),
        "ipv6-geo": _OTXParams(path="/api/v1/indicators/IPv6/{observable}/geo"),
        "dns": _OTXParams(path="/api/v1/indicators/domain/{observable}/general"),
        "dns-passivedns": _OTXParams(
            path="/api/v1/indicators/domain/{observable}/passive_dns"
        ),
        "dns-geo": _OTXParams(path="/api/v1/indicators/domain/{observable}/geo"),
        "hostname": _OTXParams(path="/api/v1/indicators/hostname/{observable}/general"),
        "file_hash": _OTXParams(path="/api/v1/indicators/file/{observable}/general"),
        "url": _OTXParams(path="/api/v1/indicators/url/{observable}/general"),
    }

    # pylint: disable=duplicate-code
    # aliases
    _IOC_QUERIES["md5_hash"] = _IOC_QUERIES["file_hash"]
    _IOC_QUERIES["sha1_hash"] = _IOC_QUERIES["file_hash"]
    _IOC_QUERIES["sha256_hash"] = _IOC_QUERIES["file_hash"]

    _REQUIRED_PARAMS = ["API_KEY"]

    def __init__(self, **kwargs):
        """Set OTX specific settings."""
        super().__init__(**kwargs)
        self.require_url_encoding = True

    def parse_results(self, response: LookupResult) -> Tuple[bool, TISeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        Tuple[bool, TISeverity, Any]
            bool = positive or negative hit
            TISeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(response.raw_result, dict):
            return False, TISeverity.information, "Not found."
        if "pulse_info" in response.raw_result:
            pulses = response.raw_result["pulse_info"].get("pulses", {})
            pulse_count = len(pulses)
            if pulse_count == 0:
                severity = TISeverity.information
                return (
                    True,
                    severity,
                    {
                        "pulse_count": pulse_count,
                        "sections_available": response.raw_result["sections"],
                    },
                )
            if pulse_count == 1:
                severity = TISeverity.warning
            else:
                severity = TISeverity.high
            return (
                True,
                severity,
                {
                    "pulse_count": pulse_count,
                    "names": [p.get("name") for p in pulses],
                    "tags": [p.get("tags") for p in pulses],
                    "references": [p.get("references") for p in pulses],
                },
            )
        return True, TISeverity.information, {}

    # pylint: enable=duplicate-code



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""HostLogonSession Entity class."""
from datetime import datetime
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .account import Account
from .host import Host

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class HostLogonSession(Entity):
    """
    HostLogonSession Entity class.

    Attributes
    ----------
    Account : Account
        HostLogonSession Account
    StartTimeUtc : datetime
        HostLogonSession StartTimeUtc
    EndTimeUtc : datetime
        HostLogonSession EndTimeUtc
    Host : Host
        HostLogonSession Host
    SessionId : str
        HostLogonSession SessionId

    """

    ID_PROPERTIES = ["Account", "Host", "SessionId"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.


        """
        self.Account: Optional[Account] = None
        self.StartTimeUtc: datetime = datetime.min
        self.EndTimeUtc: datetime = datetime.min
        self.Host: Optional[Host] = None
        self.SessionId: str = ""
        super().__init__(src_entity=src_entity, **kwargs)

        if src_event is not None:
            if "TimeCreatedUtc" in src_event:
                self.StartTimeUtc = src_event["TimeCreatedUtc"]
            elif "TimeGenerated" in src_event:
                self.StartTimeUtc = src_event["TimeGenerated"]
            self.EndTimeUtc = self.StartTimeUtc
            self.SessionId = (
                src_event["TargetLogonId"] if "TargetLogonId" in src_event else None
            )

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        if self.Host:
            return f"{self.Host.HostName}: session: {self.SessionId}"
        return self.__class__.__name__

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        if self.Host:
            return f"{self.Host.HostName}"
        return self.__class__.__name__

    _entity_schema = {
        # Account
        "Account": "Account",
        # StartTimeUtc (type System.Nullable`1[System.DateTime])
        "StartTimeUtc": None,
        # EndTimeUtc (type System.Nullable`1[System.DateTime])
        "EndTimeUtc": None,
        # Host
        "Host": "Host",
        # SessionId (type System.String)
        "SessionId": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
User configuration functions.

Loads providers based on user_defaults section in msticpyconfig.yaml

.. parsed-literal::

    UserDefaults:
    # List of query providers to load
        QueryProviders:
            AzureSentinel:
            Default:          # name of the provider listed in AzureSentinel.Workspaces
                alias: azsent   # optional - create "qry_azsent" object in globals
            CyberSoc:
                alias: soc
                connect: False  # optional - do not connect on load
            Splunk:             # add non-sentinel providers like this
            connect: False
            LocalData: local

    # List of other providers/components to load
    LoadComponents:
        TILookup:           # No parameters
        GeoIpLookup:
        provider: GeoLiteLookup   # geoip provider to use
        Notebooklets:       # Load and intialize Notebooklets
        query_provider:   # Pass it this query provider at startup
            AzureSentinel:
            workspace: CyberSoc
        Pivot:              # No parameters
        AzureData:          # auth_methods passed as startup param
        auth_methods: ['cli','interactive']
        AzureSentinelAPI:
        auth_methods: ['env','interactive']
        connect: False   # Load but do not connect


Note: For components that require authentication the default
is to connect after loading. You can skip the connect step by
add connect: False to the entry.
"""
import textwrap
from contextlib import redirect_stdout
from io import StringIO
from typing import Any, Dict, Tuple

from .._version import VERSION
from ..common.pkg_config import settings
from ..common.wsconfig import WorkspaceConfig
from ..data.data_providers import QueryProvider

__version__ = VERSION
__author__ = "Ian Hellen"


def load_user_defaults() -> Dict[str, object]:
    """
    Load providers from user defaults in msticpyconfig.yaml.

    Returns
    -------
    Dict[str, object]
        Dict of object name and provider instances.

    """
    user_defaults = settings.get("UserDefaults")
    if not user_defaults:
        return {}
    prov_dict = _load_query_providers(user_defaults)
    prov_dict.update(_load_components(user_defaults, namespace=prov_dict))

    return prov_dict


def _load_query_providers(user_defaults):
    prov_dict = {}
    query_provs = user_defaults.get("QueryProviders", {})
    if query_provs and isinstance(query_provs, dict):
        for prov_name, qry_prov_entry in user_defaults.get("QueryProviders").items():
            if prov_name == "AzureSentinel":
                provs = _load_az_workspaces(prov_name, qry_prov_entry)
                prov_dict.update(provs)
            else:
                obj_name, prov_obj = _load_provider(prov_name, qry_prov_entry)
                prov_dict[obj_name] = prov_obj
    return prov_dict


def _load_components(user_defaults, namespace=None):
    prov_dict = {}
    if "LoadComponents" not in user_defaults:
        return prov_dict
    comps_to_load = user_defaults.get("LoadComponents", {})

    for comp in COMP_LOADERS:
        load_comp_func = COMP_LOADERS.get(comp)
        if comp not in comps_to_load or not load_comp_func:
            continue
        comp_settings = comps_to_load.get(comp)

        if load_comp_func:
            print(f"Loading *{comp}*. ", end="")
            comp_out = StringIO()
            with redirect_stdout(comp_out):
                # We're calling the load_component function defined for each comp
                # and passing the dict of currently loaded providers
                # plus any global namespace passed to us (usually globals())
                obj_name, comp_obj = load_comp_func(
                    comp_settings, local_ns=prov_dict, global_ns=namespace
                )

            prov_dict[obj_name] = comp_obj
            print(f"- as '{obj_name}'")
            if comp_out.getvalue():
                print("   Component output:")
                print(
                    "\n".join(
                        textwrap.wrap(
                            comp_out.getvalue(),
                            replace_whitespace=True,
                            initial_indent="    ",
                            subsequent_indent="    ",
                        )
                    )
                )
        print()
    return prov_dict


def _load_az_workspaces(
    prov_name: str, azsent_prov_entry: Dict[str, Any]
) -> Dict[str, Any]:
    az_provs = {}
    for ws_name, ws_settings in azsent_prov_entry.items():
        print(f"Loading {prov_name}, workspace: {ws_name}")
        alias = ws_settings.get("alias", ws_name)
        connect = ws_settings.get("connect", True)
        obj_name = f"qry_{alias.lower()}"
        print(f"Workspace *{ws_name}* query provider loaded as '{obj_name}'. ", end="")
        prov_obj = QueryProvider(prov_name)

        if connect:
            ws_params = {}
            if ws_name != "Default":
                ws_params = {"workspace": ws_name}
            ws_config = WorkspaceConfig(**ws_params)  # type: ignore
            prov_obj.connect(ws_config.code_connect_str)
            print("Connected.")
        az_provs[obj_name] = prov_obj
    return az_provs


def _load_provider(prov_name: str, qry_prov_entry: Dict[str, Any]) -> Tuple[str, Any]:
    alias = qry_prov_entry.get("alias", prov_name)
    connect = qry_prov_entry.get("connect", True)
    obj_name = f"qry_{alias.lower()}"
    prov_obj = QueryProvider(prov_name)
    print(f"Loaded *{prov_name}* as '{obj_name}'. ", end="")
    if connect:
        prov_obj.connect()
        print("Connected.")
    return obj_name, prov_obj


# pylint: disable=import-outside-toplevel
def _load_ti_lookup(comp_settings=None, **kwargs):
    del comp_settings, kwargs
    from ..sectools.tilookup import TILookup

    return "ti_lookup", TILookup()


def _load_geoip_lookup(comp_settings=None, **kwargs):
    del kwargs
    provider = (
        comp_settings.get("provider") if isinstance(comp_settings, dict) else None
    )
    if provider == "GeoLiteLookup":
        from ..sectools.geoip import GeoLiteLookup

        return "geoip", GeoLiteLookup()
    if provider == "IpStackLookup":
        from ..sectools.geoip import IPStackLookup

        return "geoip", IPStackLookup()
    return None, None


def _load_notebooklets(comp_settings=None, **kwargs):
    nbinit_params = {}
    if comp_settings and isinstance(comp_settings, dict):
        prov_name, prov_args = next(
            iter(comp_settings.get("query_provider", {}).items())
        )
        if prov_name:
            nbinit_params = {"query_provider": prov_name}
        if prov_args:
            nbinit_params.update(
                {f"{prov_name}_{name}": val for name, val in prov_args.items()}
            )
    namespace = kwargs.pop("global_ns", {})
    namespace.update(kwargs.pop("local_ns", {}))
    providers = _get_provider_names(namespace)
    # Add these as additional providers
    providers = [f"+{prov}" for prov in providers]
    nbinit_params.update({"providers": providers, "namespace": namespace})
    try:
        import msticnb

        msticnb.init(**nbinit_params)
        return "nb", msticnb
    except ImportError:
        print("Cannot load MSTIC notebooklets (msticnb) package.")
        print("Please install - 'pip install msticnb'")

    return None, None


def _load_pivot(comp_settings=None, **kwargs):
    del comp_settings
    from ..datamodel.pivot import Pivot

    namespace = kwargs.get("global_ns", {}).copy()
    namespace.update(kwargs.get("local_ns", {}))
    piv_kwargs = {"namespace": namespace}
    pivot = Pivot(**piv_kwargs)
    return "pivot", pivot


def _load_azure_data(comp_settings=None, **kwargs):
    del kwargs
    from ..data.azure_data import AzureData

    az_data = AzureData()
    connect = comp_settings.pop("connect", True)
    connect_args = comp_settings
    if connect:
        az_data.connect(**connect_args)
        print("Connected. ", end="")
    return "az_data", az_data


def _load_azsent_api(comp_settings=None, **kwargs):
    del kwargs
    from ..data.azure_sentinel import AzureSentinel

    az_sent = AzureSentinel()
    connect = comp_settings.pop("connect", True)
    connect_args = comp_settings
    if connect:
        az_sent.connect(**connect_args)
        print("Connected. ", end="")
    return "azs_api", az_sent


# providers loaded in order specified
# Do not alter the order unless you know what dependencies
# are between the different providers.
COMP_LOADERS = {
    "TILookup": _load_ti_lookup,
    "GeoIpLookup": _load_geoip_lookup,
    "AzureData": _load_azure_data,
    "AzureSentinelAPI": _load_azsent_api,
    "Pivot": _load_pivot,  # Pivots loaded after most providers
    "Notebooklets": _load_notebooklets,  # Notebooklets calls add_pivots
}


def _get_provider_names(prov_dict):
    providers = []
    for _, obj in prov_dict.items():
        if isinstance(obj, QueryProvider) and obj.connected:
            providers.append(obj.environment.casefold())
        else:
            cls_name = obj.__class__.__name__
            if cls_name in COMP_LOADERS and cls_name != "Pivot":
                providers.append(cls_name.casefold())
    return providers



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data driver base class."""
import abc
from abc import ABC
from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, Optional, Set, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ..query_source import QuerySource

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-instance-attributes
class DriverBase(ABC):
    """Base class for data providers."""

    def __init__(self, **kwargs):
        """Initialize new instance."""
        self._kwargs = kwargs
        self._loaded = False
        self._connected = False
        self.current_connection = None
        self.public_attribs: Dict[str, Any] = {}
        self.formatters: Dict[str, Callable] = {}
        self.use_query_paths = True
        self.has_driver_queries = False
        self._previous_connection = False
        self.data_environment = kwargs.get("data_environment")
        self._query_filter: Dict[str, Set[str]] = defaultdict(set)

    @property
    def loaded(self) -> bool:
        """
        Return true if the provider is loaded.

        Returns
        -------
        bool
            True if the provider is loaded.

        Notes
        -----
        This is not relevant for some providers.

        """
        return self._loaded

    @property
    def connected(self) -> bool:
        """
        Return true if at least one connection has been made.

        Returns
        -------
        bool
            True if a successful connection has been made.

        Notes
        -----
        This does not guarantee that the last data source
        connection was successful. It is a best effort to track
        whether the provider has made at least one successful
        authentication.

        """
        return self._connected

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return {}

    @abc.abstractmethod
    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : Optional[str]
            Connect to a data source

        """
        return None

    @abc.abstractmethod
    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        kwargs :
            Are passed to the underlying provider query method,
            if supported.

        Returns
        -------
        Union[pd.DataFrame, Any]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """

    @abc.abstractmethod
    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame plus native results.

        Parameters
        ----------
        query : str
            The query to execute

        Returns
        -------
        Tuple[pd.DataFrame,Any]
            A DataFrame and native results.

        """

    @property
    def service_queries(self) -> Tuple[Dict[str, str], str]:
        """
        Return queries retrieved from the service after connecting.

        Returns
        -------
        Tuple[Dict[str, str], str]
            Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        return {}, ""

    @property
    def driver_queries(self) -> Iterable[Dict[str, Any]]:
        """
        Return queries retrieved from the service after connecting.

        Returns
        -------
        List[Dict[str, str]]
            List of Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        return [{}]

    @property
    def query_attach_spec(self) -> Dict[str, Set[str]]:
        """Parameters that determine whether a query is relevant for the driver."""
        return self._query_filter

    def add_query_filter(self, name, query_filter):
        """Add an expression to the query attach filter."""
        allowed_names = {"data_environments", "data_families", "data_sources"}
        if name not in allowed_names:
            raise ValueError(
                f"'name' {name} must be one of:",
                ", ".join(f"'{name}'" for name in allowed_names),
            )
        self._query_filter[name].add(query_filter)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for creating sessions out of raw data."""

from typing import List

import numpy as np
import pandas as pd
from pandas.core.dtypes.dtypes import DatetimeTZDtype


def sessionize_data(
    data: pd.DataFrame,
    user_identifier_cols: List[str],
    time_col: str,
    max_session_time_mins: int,
    max_event_separation_mins: int,
    event_col: str,
) -> pd.DataFrame:
    """
    Sessionize the input data.

    In particular, the resulting dataframe will have 1 row per session. It will contain the
    following columns: the user_identifier_cols, <time_col>_min, <time_col>_max,
    <event_col>_list, duration (<time_col>_max - <time_col>_min), number_events (length of the
    <event_col>_list value)

    Parameters
    ----------
    data: pd.DataFrame
        This dataframe should contain at least the following columns:
        - time stamp column
        - columns related to user name and/or computer name and/or ip address etc
        - column containing an event
    user_identifier_cols: List[str]
        Name of the columns which contain username and/or computer name and/or ip address etc.
        Each time the value of one of these columns changes, a new session will be started.
    time_col: str
        Name of the column which contains a time stamp.
        If this column is not already in datetime64[ns, UTC] format, it will be casted to it.
    max_session_time_mins: int
        The maximum length of a session in minutes. If a sequence of events for the same
        user_identifier_cols values exceeds this length, then a new session will be started.
    max_event_separation_mins: int
        The maximum length in minutes between two events in a session. If we have 2 events for
        the same user_identifier_cols values, and if those two events are more than
        `max_event_separation_mins` apart, then a new session will be started.
    event_col: str
        Name of the column which contains the event of interest.
        For example, if we are interested in sessionizing exchange admin commands,
        the "event_col" could contain values like: "Set-Mailbox" or "Set-User" etc.

    Returns
    -------
    pd.DataFrame containing the sessionized data. 1 row per session.

    """
    df_with_sesind = create_session_col(
        data=data,
        user_identifier_cols=user_identifier_cols,
        time_col=time_col,
        max_session_time_mins=max_session_time_mins,
        max_event_separation_mins=max_event_separation_mins,
    )

    # aggregating will not work properly with nans. Temporarily replace nan values with dummy_str.
    for col in user_identifier_cols:
        df_with_sesind[col] = df_with_sesind[col].fillna("dummy_str")

    # aggregate by the session_ind column
    agg_df = (
        df_with_sesind.sort_values(["session_ind", time_col])
        .groupby(["session_ind"] + user_identifier_cols, as_index=False)
        .agg({time_col: ["min", "max"], event_col: list})
        .reset_index()
    )

    # rename some columns
    outer_cols = agg_df.columns.get_level_values(0)
    inner_cols = agg_df.columns.get_level_values(1)
    agg_df.columns = [
        outer_cols[i] + "_" + inner_cols[i] if inner_cols[i] != "" else outer_cols[i]
        for i in range(len(outer_cols))
    ]

    # calculate some additional columns
    agg_df["duration"] = agg_df[f"{time_col}_max"] - agg_df[f"{time_col}_min"]
    agg_df["number_events"] = agg_df[f"{event_col}_list"].apply(len)

    agg_df = agg_df.drop("session_ind", axis=1)
    if "index" in agg_df.columns:
        agg_df = agg_df.drop("index", axis=1)

    # replace dummy_str with nan values
    for col in user_identifier_cols:
        agg_df[col] = agg_df[col].replace("dummy_str", np.nan)

    return agg_df


# pylint: disable=too-many-locals, too-many-branches
def create_session_col(
    data: pd.DataFrame,
    user_identifier_cols: List[str],
    time_col: str,
    max_session_time_mins: int,
    max_event_separation_mins: int,
) -> pd.DataFrame:
    """
    Create a "session_ind" column in the dataframe.

    In particular, the session_ind column will be incremented each time a new session
    starts.

    Parameters
    ----------
    data: pd.DataFrame
        This dataframe should contain at least the following columns:
        - time stamp column
        - columns related to user name and/or computer name and/or ip address etc
    user_identifier_cols: List[str]
        Name of the columns which contain username and/or computer name and/or ip address etc.
        Each time the value of one of these columns changes, a new session will be started.
    time_col: str
        Name of the column which contains a time stamp.
        If this column is not already in datetime64[ns, UTC] format, it will be casted to it.
    max_session_time_mins: int
        The maximum length of a session in minutes. If a sequence of events for the same
        user_identifier_cols values exceeds this length, then a new session will be started.
    max_event_separation_mins: int
        The maximum length in minutes between two events in a session. If we have 2 events for
        the same user_identifier_cols values, and if those two events are more than
        `max_event_separation_mins` apart, then a new session will be started.

    Returns
    -------
    pd.DataFrame with an additional "session_ind" column

    """
    max_sep = pd.to_timedelta(max_event_separation_mins, unit="min")
    max_ses = pd.to_timedelta(max_session_time_mins, unit="min")

    df_with_sesind = data.copy()
    if not isinstance(df_with_sesind[time_col].dtype, DatetimeTZDtype):
        df_with_sesind[time_col] = pd.to_datetime(df_with_sesind[time_col])

    final_cols = list(df_with_sesind.columns) + ["session_ind"]

    if len(df_with_sesind) == 0:
        df_with_sesind["session_ind"] = None
        return df_with_sesind

    # Sessionising will not work properly with nans. Temporarily replace nan values with dummy_str.
    for col in user_identifier_cols:
        df_with_sesind[col] = df_with_sesind[col].fillna("dummy_str")

    df_with_sesind = df_with_sesind.sort_values(
        user_identifier_cols + [time_col]
    ).reset_index(drop=True)

    # initialise first row
    ses_ind = 0
    df_with_sesind.loc[0, "time_diff"] = pd.to_timedelta(0)
    df_with_sesind.loc[0, "cml_time"] = pd.to_timedelta(0)
    df_with_sesind.loc[0, "session_ind"] = ses_ind

    for i in range(1, len(df_with_sesind)):
        cur = df_with_sesind.iloc[i]
        prev = df_with_sesind.iloc[i - 1]

        # if any of the user_identifier_cols values change, a new session should start
        new_flag = False
        for col in user_identifier_cols:
            if cur[col] != prev[col]:
                new_flag = True
                break

        dif = cur[time_col] - prev[time_col]
        cml = prev["cml_time"] + dif
        # if the max session length is exceeded or the max separation between events is exceeded,
        # a new session should start
        if dif > max_sep or cml > max_ses:
            new_flag = True

        if new_flag:
            df_with_sesind.loc[i, "time_diff"] = pd.to_timedelta(0)
            df_with_sesind.loc[i, "cml_time"] = pd.to_timedelta(0)
            ses_ind += 1
            df_with_sesind.loc[i, "session_ind"] = ses_ind
        else:
            df_with_sesind.loc[i, "time_diff"] = dif
            df_with_sesind.loc[i, "cml_time"] = cml
            df_with_sesind.loc[i, "session_ind"] = ses_ind

    # replace dummy_str with nan values
    for col in user_identifier_cols:
        df_with_sesind[col] = df_with_sesind[col].replace("dummy_str", np.nan)

    return df_with_sesind[final_cols]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MP Config Control Class."""
from collections import namedtuple
import pkgutil
import re
from typing import Any, Dict, List, Optional, Tuple, Union

import ipywidgets as widgets
import yaml

from ..common.utility import is_valid_uuid
from .._version import VERSION
from .ce_common import print_debug, py_to_widget, widget_to_py
from .comp_edit import SettingsControl

__version__ = VERSION
__author__ = "Ian Hellen"


STORE_TEXT = "Text"
STORE_ENV_VAR = "EnvironmentVar"
STORE_KEYVAULT = "KeyVault"
STORE_OPT = "StoreType"

ValidtnResult = namedtuple("ValidtnResult", "result, status")
_VALID_SUCESS = "Validation succeeded"


class MpConfigControls:
    """Msticpy configuration and settings database."""

    def __init__(
        self, mp_config_def: Dict[str, Any], mp_config: Optional[Dict[str, Any]] = None
    ):
        """
        Return an instance of MpConfigControls.

        Parameters
        ----------
        mp_config_def : Dict[str, Any]
            Msticpy config setting definitions.
        mp_config : Optional[Dict[str, Any]], optional
            Msticpy Settings dictionary, by default None

        """
        self.mp_config = mp_config or {}
        self._raw_config_defn = mp_config_def
        self.config_defn = self._convert_mp_config(mp_config_def)
        self.controls = self._create_ctrl_dict(self.mp_config)

        if "AzureCLI" in self.mp_config:
            if "DataProviders" not in self.mp_config:
                self.mp_config["DataProviders"] = {}
            self.mp_config["DataProviders"]["AzureCLI"] = self.mp_config["AzureCLI"]

    @staticmethod
    def _get_elem_from_path(path, member_dict: Dict[str, Any]):
        """Return an item at the path from `member_dict`."""
        paths = path.split(".")
        current_elem = member_dict
        for elem in paths:
            next_elem = current_elem.get(elem)
            if next_elem is None:
                return next_elem
            current_elem = next_elem
        return current_elem

    def _set_elem_at_path(
        self, path: str, member_dict: Dict[str, Any], value: Any, create: bool = True
    ):
        """Set item at the path from `member_dict` to `value`."""
        path_elems = path.rsplit(".", maxsplit=1)
        parent_path = path_elems[0]
        parent = self._get_elem_from_path(parent_path, member_dict)
        if parent is not None:
            if len(path_elems) > 1:
                parent[path_elems[1]] = value
            else:
                member_dict[parent_path] = value
        elif create:
            # if an element in the path doesn't exist, create any needed
            if len(path_elems) == 1:
                member_dict[parent_path] = value
                return
            tgt_key = path_elems[1]
            current_elem = member_dict
            for elem in parent_path.split("."):
                next_elem = current_elem.get(elem)
                if next_elem is None:
                    next_elem = {}
                    current_elem[elem] = next_elem
                current_elem = next_elem
            current_elem[tgt_key] = value
            print_debug("set", parent_path, tgt_key, value)

    def _del_elem_at_path(self, path: str, member_dict: Dict[str, Any]):
        """Delete an item at `path`."""
        parent_path, tgt_key = path.rsplit(".", maxsplit=1)
        parent = self._get_elem_from_path(parent_path, member_dict)
        if parent is not None and tgt_key in parent:
            del parent[tgt_key]

    def get_value(self, path: str) -> Any:
        """Return setting value at `path`."""
        return self._get_elem_from_path(path, self.mp_config)

    def set_value(self, path: str, value: Any):
        """Set setting value at path to `value`."""
        self._set_elem_at_path(path, self.mp_config, value)

    def del_value(self, path: str, keep_ctrl: bool = False):
        """Delete setting item at `path`."""
        self._del_elem_at_path(path, self.mp_config)
        if not keep_ctrl:
            self.del_control(path)

    def get_control(self, path: str) -> Any:
        """Return the control stored at `path`."""
        return self._get_elem_from_path(path, self.controls)

    def set_control(self, path: str, control):
        """Set the control stored at `path`."""
        print_debug("set_control", path)
        self._set_elem_at_path(path, self.controls, value=control)

    def del_control(self, path: str):
        """Delete the control stored at `path`."""
        self._del_elem_at_path(path, self.controls)

    def get_defn(self, path: str) -> Union[Dict[str, Any], Tuple[str, Any]]:
        """Return the setting definition at `path`."""
        defn = self._get_elem_from_path(path, self.config_defn)
        if defn is not None:
            return defn
        if path.startswith("AzureSentinel.Workspaces"):
            path = re.sub(
                r"(?P<root>AzureSentinel\.Workspaces\.)([^.]+)", r"\1Default", path
            )
        return self._get_elem_from_path(path, self.config_defn)

    def rename_path(self, old_path: str, new_path: str):
        """Rename a setting from `old_path` to `new_path`."""
        old_path_elems = old_path.rsplit(".", maxsplit=1)
        new_path_elems = new_path.rsplit(".", maxsplit=1)
        if (
            old_path_elems[0] != new_path_elems[0]
            or len(old_path_elems) == 1
            or len(new_path_elems) == 1
        ):
            raise ValueError(
                "Can only rename the bottom element of paths", old_path, new_path
            )
        path_root = old_path_elems[0]
        src_key = old_path_elems[1]
        tgt_key = new_path_elems[1]
        for setting_dict in (self.mp_config, self.controls):
            parent_elem = self._get_elem_from_path(path_root, setting_dict)
            parent_elem[tgt_key] = parent_elem.get(src_key)
            if src_key in parent_elem:
                del parent_elem[src_key]

    def populate_ctrl_values(self, path: str):
        """Populate control at `path` from settings at `path`."""
        print_debug("populate_ctrl_values", path)
        self._set_ctrl_values(path)

    def _set_ctrl_values(self, path: str):
        """Recursively set settings from control values."""
        ctrl_tree = self.get_control(path)
        if isinstance(ctrl_tree, widgets.Widget):
            print_debug("_set_ctrl_values - widget", ctrl_tree)
            curr_value = py_to_widget(self.get_value(path), ctrl=ctrl_tree)
            if curr_value:
                ctrl_tree.value = curr_value
        elif isinstance(ctrl_tree, SettingsControl):
            print_debug(
                "_set_ctrl_values - _SettingsControl",
                ctrl_tree,
                ctrl_tree.value,
                path,
                self.get_value(path),
            )
            curr_value = self.get_value(path)
            if curr_value:
                ctrl_tree.value = curr_value
        elif isinstance(ctrl_tree, dict):
            for key in ctrl_tree:
                self._set_ctrl_values(f"{path}.{key}")

    def save_ctrl_values(self, path: str):
        """Save the values in the control at `path` to settings."""
        self.set_value(path, self._get_ctrl_values(path))

    def _get_ctrl_values(self, path: str):
        """Recursively save values in the control at `path` to settings."""
        ctrl_tree = self.get_control(path)
        print_debug(
            type(ctrl_tree),
            "instance check",
            isinstance(ctrl_tree, (widgets.Widget, SettingsControl)),
        )
        if isinstance(ctrl_tree, (widgets.Widget, SettingsControl)):
            return widget_to_py(ctrl_tree)
        if isinstance(ctrl_tree, dict):
            return {key: self._get_ctrl_values(f"{path}.{key}") for key in ctrl_tree}
        raise ValueError(f"No control found at path {path}")

    def _create_ctrl_dict(self, config_dict):
        """Create a blank control dictionary from settings."""
        ctrl_dict = config_dict.copy()
        for name, val in ctrl_dict.items():

            if (
                isinstance(val, dict)
                and STORE_KEYVAULT not in val
                and STORE_ENV_VAR not in val
            ):
                ctrl_dict[name] = self._create_ctrl_dict(val)
            else:
                ctrl_dict[name] = None
        return ctrl_dict

    def validate_all_settings(self, show_all: bool = False) -> List[ValidtnResult]:
        """
        Validate settings against definitions.

        Parameters
        ----------
        show_all : bool, optional
            Show success validations as well as failures, by default False

        Returns
        -------
        List[ValidtnResult]
            List of validation results:
            bool - True == valid
            status - validation result description

        """
        results = []
        for key in self.mp_config:
            results.extend(self.validate_setting(path=key, show_all=show_all))
        return results

    def validate_setting(
        self, path: str, defn_path: Optional[str] = None, show_all: bool = False
    ) -> List[ValidtnResult]:
        """
        Validate settings against definitions for a specific path.

        Parameters
        ----------
        path : str
            The setting path
        defn_path : Optional[str], optional
            The definition path, by default None
            Unless, specified this is the same as the setting path
        show_all : bool, optional
            Return successful as well as failed validation results
            , by default False

        Returns
        -------
        List[ValidtnResult]
            List of validation results:
            bool - True == valid
            status - validation result description

        """
        results = self._validate_setting_at_path(path=path, defn_path=defn_path)
        if isinstance(results, tuple):
            return [results] if show_all or not results[0] else []
        if isinstance(results, list):
            up_results = self._unpack_lists(results)
            return [res for res in up_results if not res[0] or show_all]
        return [ValidtnResult(True, "No validation results found")]

    def _unpack_lists(self, res_list: List[Any]) -> List[ValidtnResult]:
        """Unpack nested lists into a single list."""
        results = []
        for item in res_list:
            if isinstance(item, list):
                results.extend(self._unpack_lists(item))
            else:
                results.append(item)
        return results

    # pylint: disable=too-many-return-statements
    def _validate_setting_at_path(
        self, path: str, defn_path: Optional[str] = None, index: Optional[int] = None
    ) -> Union[ValidtnResult, List[Union[ValidtnResult, List[Any]]]]:
        """Recursively validate settings at path."""
        defn_path = defn_path or path
        conf_defn = self.get_defn(defn_path)
        setting = self.get_value(path)
        # to support recursive calls on members of lists (in the
        # if isinstance(conf_defn, list) section) we need to pull out
        # which item we're testing here
        if index is not None and setting is not None:
            setting = setting[index]

        # If we don't have a definition, there's nothing to validate against.
        if not conf_defn:
            return ValidtnResult(True, f"No definition for path '{path}'")

        # If this is None and we're not at a leaf setting it means the
        # settings path doesn't exist, so bail here.
        if not setting and not isinstance(conf_defn, tuple):
            return ValidtnResult(True, f"No setting at path '{path}'")

        # if the current defn node is a dict, then we need to recurse
        # into it.
        if isinstance(conf_defn, dict):
            val_res = [
                self._validate_setting_at_path(
                    path=f"{path}.{key}", defn_path=f"{defn_path}.{key}"
                )
                for key in conf_defn
            ]
            return val_res[0] if len(val_res) == 1 else val_res

        # if the defn is a list, we need to iterate through the members
        if isinstance(conf_defn, list):
            if isinstance(setting, list):
                # if the setting value is a list - check each one
                val_res = [
                    self._validate_setting_at_path(path, defn_path, index=idx)
                    for idx, _ in enumerate(setting)
                ]
                return val_res[0] if len(val_res) == 1 else val_res
            # Otherwise assume that we are already checking an item in
            # the list (recusing) - so we need to pull the defn from the
            # first item in the definition list
            conf_defn = conf_defn[0]

        # If the defn is a tuple we've reached a leaf node with a true
        # definition
        if isinstance(conf_defn, tuple):
            val_type, val_opts = conf_defn
            validator = _VALIDATORS.get(val_type)
            if validator:
                return validator(setting, path, val_type, val_opts)
            return ValidtnResult(True, "No validator for type '{val_type}'")

        # If the conf definition is a literal value - compare directly
        if conf_defn == setting:
            return ValidtnResult(True, f"Value is valid at path '{path}'")

        # Otherwise, we assume failure.
        return ValidtnResult(False, f"Validation failed for path '{path}'")

    def _yml_extract_type(self, conf_val):
        """Extract type and options from definition."""
        if not conf_val or "(" not in conf_val or ")" not in conf_val:
            return "unknown", {}
        val_type_match = re.match(
            r"(?P<type>[^()]+)\((?P<params>.*)\)$", conf_val.strip()
        )
        val_type = val_type_match.groupdict().get("type")
        val_param_str = val_type_match.groupdict().get("params", "")

        if val_param_str:
            val_params = {
                param.split("=")[0].strip(): param.split("=")[1].strip()
                for param in val_param_str.split(",")
                if "=" in param
            }
            val_params = {
                key: True if val == "True" else False if val == "False" else val
                for key, val in val_params.items()
            }
        else:
            val_params = {}
        if "options" in val_params:
            val_params["options"] = [
                val.strip("'\"")
                for val in val_params["options"].strip()[1:-1].split("; ")
            ]
        if "mp_defn_path" in val_params:
            defn_path = val_params.pop("mp_defn_path").strip(" /\"'").replace("/", ".")
            defn = self._get_elem_from_path(defn_path, self._raw_config_defn)
            val_params["defn"] = defn
        default_val = val_params.get("default", "")
        if isinstance(default_val, str):
            if default_val.strip().startswith("["):
                # This is a default list - so we need to parse it
                val_params["default"] = [
                    val.strip("'\"")
                    for val in re.split(r"\s*;\s*|\s*,\s*", default_val.strip()[1:-1])
                ]
            else:
                val_params["default"] = default_val.strip("'\"")
        return val_type, val_params

    def _convert_mp_config(self, mp_conf_item):
        """Convert definition dictionary to extract definitions."""
        if isinstance(mp_conf_item, list):
            return self._convert_mp_config_list(mp_conf_item)
        if isinstance(mp_conf_item, dict):
            return self._convert_mp_config_dict(mp_conf_item)
        return mp_conf_item

    def _convert_mp_config_dict(self, mp_conf_dict):
        """Recursively convert definition dictionary to extract definitions."""
        out_dict = {}
        for key, val in mp_conf_dict.items():
            if isinstance(val, dict):
                out_dict[key] = self._convert_mp_config_dict(val)
            elif isinstance(val, list):
                out_dict[key] = self._convert_mp_config_list(val)
            else:
                key_type, opts = self._yml_extract_type(val)
                out_dict[key] = val if key_type == "unknown" else (key_type, opts)
        return out_dict

    def _convert_mp_config_list(self, mp_conf_list):
        """Recursively convert definition list to extract definitions."""
        out_list = []
        for val in mp_conf_list:
            if isinstance(val, dict):
                out_list.append(self._convert_mp_config_dict(val))
            elif isinstance(val, list):
                out_list.append(self._convert_mp_config_list(val))
            else:
                key_type, opts = self._yml_extract_type(val)
                out_list.append(val if key_type == "unknown" else (key_type, opts))
        return out_list


def get_mpconfig_definitions() -> Dict[str, Any]:
    """
    Return the current msticpyconfig definition dictionary.

    Returns
    -------
    Dict[str, Any]
        msticpyconfig definition dictionary

    Raises
    ------
    ValueError:
        Could not load definitions from resources/mpconfig_defaults.yaml

    """
    pkg_root = __package__.split(".", maxsplit=1)[0]
    file_bytes = pkgutil.get_data(pkg_root, "resources/mpconfig_defaults.yaml")
    if file_bytes:
        return yaml.safe_load(file_bytes)
    raise ValueError("Could not load definitions from resources/mpconfig_defaults.yaml")


# Supporting functions for MpConfigControls
def _get_mssg(value, path):
    return f"value '{value}', for setting at path '{path}'"


def _is_none_and_not_required(value, val_type, val_opts):
    if value is None or (val_type == "str" and not value):
        return not val_opts.get("required", True)
    return False


def _validate_string(value, path, val_type, val_opts):
    mssg = _get_mssg(value, path)
    if _is_none_and_not_required(value, val_type, val_opts):
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    if not isinstance(value, str):
        return ValidtnResult(
            False, f"Value type {type(value)} should be type {val_type} - {mssg}"
        )
    if "options" in val_opts and value not in val_opts["options"]:
        return ValidtnResult(
            False,
            f"Value {value} must be one of {', '.join(val_opts['options'])} - {mssg}",
        )
    if val_opts.get("format") == "uuid" and not is_valid_uuid(value):
        return ValidtnResult(
            False,
            f"Value {value} should be a UUID - {mssg}",
        )
    return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")


def _validate_bool(value, path, val_type, val_opts):
    mssg = _get_mssg(value, path)
    if _is_none_and_not_required(value, val_type, val_opts):
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    if not isinstance(value, bool):
        return ValidtnResult(
            False, f"Value type {type(value)} should be type {val_type} - {mssg}"
        )
    return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")


def _validate_m_enum(value, path, val_type, val_opts):
    mssg = _get_mssg(value, path)
    if _is_none_and_not_required(value, val_type, val_opts):
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    if not isinstance(value, (str, list)):
        return ValidtnResult(
            False, f"Value type {type(value)} should be type {val_type} - {mssg}"
        )
    if "options" in val_opts:
        if isinstance(value, str) and value not in val_opts["options"]:
            return ValidtnResult(
                False,
                f"Value {value} must be one of {', '.join(val_opts['options'])} - {mssg}",
            )
        if not isinstance(value, list):
            return ValidtnResult(
                False,
                f"Value '{value}' should be a string or list. "
                + f"Must be one of {', '.join(val_opts['options'])} - {mssg}",
            )
        invalid_opts = [val for val in value if val not in val_opts["options"]]
        if invalid_opts:
            return ValidtnResult(
                False,
                f"Invalid values '{invalid_opts}' found. "
                + f"Must be one of {', '.join(val_opts['options'])} - {mssg}",
            )
    return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")


def _validate_txt_dict(value, path, val_type, val_opts):
    mssg = _get_mssg(value, path)
    if _is_none_and_not_required(value, val_type, val_opts):
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    if isinstance(value, dict):
        for d_key, d_val in value.items():
            if not isinstance(d_key, str):
                return ValidtnResult(
                    False,
                    f"Key {d_key} of {value} must be a string - {mssg}",
                )
            if not isinstance(d_val, (str, int, bool)):
                return ValidtnResult(
                    False,
                    f"Value {d_val} of key {d_key} in {value} must be a"
                    + f" string, int or bool - {mssg}",
                )
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    return ValidtnResult(False, f"Value {value} should be a dictionary - {mssg}")


def _validate_list(value, path, val_type, val_opts):
    mssg = _get_mssg(value, path)
    if _is_none_and_not_required(value, val_type, val_opts):
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    if isinstance(value, list):
        for item_val in value:
            expected_type = val_opts.get("elem_type", "str")
            val_type = type(item_val).__name__
            if expected_type != val_type:
                return ValidtnResult(
                    False,
                    f"Item {item_val} of {value} expected to be a {expected_type} - {mssg}",
                )
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    return ValidtnResult(False, f"Value {value} should be a list - {mssg}")


def _validate_defn(value, path, val_type, val_opts):
    """
    Validate a compound type definition.

    We want to validate a value against a definition like:
    ::
        cred_key:
            one_of:
                - str: str()
                - EnvironmentVar: str()
                - KeyVault: str(required=False)

    This only handles this type of definition and only looks at
    the "required" flag in the options. We'll extend this checker if we
    need to for other options.

    """
    mssg = _get_mssg(value, path)
    if _is_none_and_not_required(value, val_type, val_opts):
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")

    # This only handles "one_of" lists of alternatives
    opt_list = val_opts.get("defn", {}).get("one_of")
    if not opt_list:
        return ValidtnResult(True, f"{_VALID_SUCESS} {mssg}")
    # pull the definitions (val_type and val_opts) for each item into a dict
    opt_dict = {next(iter(val.keys())): next(iter(val.values())) for val in opt_list}

    # if the checked value is a string, validate the string
    if isinstance(value, str) and "str" in opt_dict:
        opt_v_opts = {"required": "required=False" not in opt_dict["str"]}
        return _validate_string(value, path, "str", opt_v_opts)

    # If the value is a dict
    if isinstance(value, dict):
        # We assume that the value is a dict with a single key
        v_key, v_val = next(iter(value.items()))
        # If the name is in the one_of dictionary names
        if v_key in opt_dict:
            # get the element type
            opt_v_type = re.sub(r"\(.*\)", "", opt_dict[v_key])
            # add "required=False" if that is the value options for this item
            opt_v_opts = {"required": "required=False" not in opt_dict[v_key]}
            # get the validator function for the option type and use that
            # to check the actual value
            _validator = _VALIDATORS.get(opt_v_type, _validate_string)
            return _validator(v_val, f"{path}.{v_key}", opt_v_type, opt_v_opts)
    # Otherwise the validation failed
    return ValidtnResult(
        False,
        f"Value type {type(value)} does not match definition {val_opts['defn']} - {mssg}",
    )


_VALIDATORS = {
    "str": _validate_string,
    "enum": _validate_string,
    "m_enum": _validate_m_enum,
    "bool": _validate_bool,
    "cred_key": _validate_defn,
    "txt_dict": _validate_txt_dict,
    "list": _validate_list,
}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""KQL Driver class."""
import json
import os
import re
import warnings
from datetime import datetime
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import pandas as pd
from IPython import get_ipython

from ...common.azure_auth import AzureCloudConfig, only_interactive_cred, az_connect
from ...common.exceptions import (
    MsticpyDataQueryError,
    MsticpyImportExtraError,
    MsticpyKqlConnectionError,
    MsticpyNoDataSourceError,
    MsticpyNotConnectedError,
)
from ...common.wsconfig import WorkspaceConfig
from ...common.utility import export
from ..query_defns import DataEnvironment
from .driver_base import DriverBase, QuerySource

try:
    from Kqlmagic import kql as kql_exec
    from Kqlmagic.kql_engine import KqlEngineError
    from Kqlmagic.kql_response import KqlError
    from Kqlmagic.my_aad_helper import AuthenticationError
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without Kqlmagic installed",
        title="Error importing Kqlmagic",
        extra="kql",
    ) from imp_err

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_KQL_CLOUD_MAP = {
    "global": "public",
    "cn": "china",
    "usgov": "government",
    "de": "germany",
}

_AZ_CLOUD_MAP = {kql_cloud: az_cloud for az_cloud, kql_cloud in _KQL_CLOUD_MAP.items()}

_LOGANALYTICS_URL_BY_CLOUD = {
    "global": "https://api.loganalytics.io/",
    "cn": "https://api.loganalytics.azure.cn/",
    "usgov": "https://api.loganalytics.us/",
    "de": "https://api.loganalytics.de/",
}


# pylint: disable=too-many-instance-attributes


@export
class KqlDriver(DriverBase):
    """KqlDriver class to execute kql queries."""

    def __init__(self, connection_str: str = None, **kwargs):
        """
        Instantiate KqlDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        Other Parameters
        ----------------
        debug : bool
            print out additional diagnostic information.

        """
        self._ip = get_ipython()
        self._debug = kwargs.get("debug", False)
        super().__init__(**kwargs)

        self.formatters = {"datetime": self._format_datetime, "list": self._format_list}
        self._loaded = self._is_kqlmagic_loaded()

        os.environ["KQLMAGIC_LOAD_MODE"] = "silent"
        if not self._loaded:
            self._load_kql_magic()

        self._schema: Dict[str, Any] = {}
        self.environment = kwargs.get("data_environment", DataEnvironment.MSSentinel)
        self.kql_cloud, self.az_cloud = self._set_kql_cloud()

        if connection_str:
            self.current_connection = connection_str
            self.connect(connection_str)

    # pylint: disable=too-many-branches
    def connect(self, connection_str: Optional[str] = None, **kwargs):  # noqa: MC0001
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : str
            Connect to a data source

        Other Parameters
        ----------------
        kqlmagic_args : str, optional
            Additional string of parameters to be passed to KqlMagic
        mp_az_auth : Union[bool, str, list, None], optional
            Optional parameter directing KqlMagic to use MSTICPy Azure authentication.
            Values can be:
            True or "default": use the settings in msticpyconfig.yaml 'Azure' section
            str: single auth method name ('msi', 'cli', 'env' or 'interactive')
            List[str]: list of acceptable auth methods from ('msi', 'cli',
            'env' or 'interactive')

        """
        if not self._previous_connection:
            print("Connecting...", end=" ")
        if isinstance(connection_str, WorkspaceConfig):
            connection_str = connection_str.code_connect_str
        if not connection_str:
            raise MsticpyKqlConnectionError(
                f"A connection string is needed to connect to {self._connect_target}",
                title="no connection string",
            )
        if "kqlmagic_args" in kwargs:
            connection_str = connection_str + " " + kwargs["kqlmagic_args"]
        # Default to using Azure Auth if possible.
        mp_az_auth = kwargs.pop("mp_az_auth", "default")
        if mp_az_auth and "try_token" not in kwargs:
            self._set_az_auth_option(mp_az_auth)
        self.current_connection = connection_str
        kql_err_setting = self._get_kql_option("short_errors")
        self._connected = False
        try:
            self._set_kql_option("short_errors", False)
            if self._ip is not None:
                try:
                    kql_exec(connection_str)
                    if not self._previous_connection:
                        print("connected")
                except KqlError as ex:
                    self._raise_kql_error(ex)
                except KqlEngineError as ex:
                    self._raise_kql_engine_error(ex)
                except AuthenticationError as ex:
                    self._raise_authn_error(ex)
                except Exception as ex:  # pylint: disable=broad-except
                    self._raise_adal_error(ex)
                self._connected = True
                self._previous_connection = True
                self._schema = self._get_schema()
            else:
                print(f"Could not connect to kql query provider for {connection_str}")
            return self._connected
        finally:
            self._set_kql_option("short_errors", kql_err_setting)

    # pylint: disable=too-many-branches

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return self._schema

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """
        del kwargs
        if query_source:
            try:
                table = query_source["args.table"]
            except KeyError:
                table = None
            if table:
                if " " in table.strip():
                    table = table.strip().split(" ")[0]
                if table not in self.schema:
                    raise MsticpyNoDataSourceError(
                        f"The table {table} for this query is not in your workspace",
                        " or database schema. Please check your this",
                        title=f"{table} not found.",
                    )
        data, result = self.query_with_results(query)
        return data if data is not None else result

    # pylint: disable=too-many-branches
    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The kql query to execute

        Returns
        -------
        Tuple[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) and
            Kql ResultSet.

        """
        # connect or switch the connection if our connection string
        # is not the current KqlMagic connection.
        # if not self.connected and self.current_connection:
        try:
            self.connect(self.current_connection)
        except MsticpyKqlConnectionError:
            self._connected = False
        if not self.connected:
            raise MsticpyNotConnectedError(
                "Please run the connect() method before running a query.",
                title=f"not connected to a {self._connect_target}",
                help_uri=MsticpyKqlConnectionError.DEF_HELP_URI,
            )

        if self._debug:
            print(query)

        # save current auto_dataframe setting so that we can set to false
        # and restore current setting
        auto_dataframe = self._get_kql_option(option="auto_dataframe")
        self._set_kql_option(option="auto_dataframe", value=False)

        # run the query (append semicolon to prevent default output)
        if not query.strip().endswith(";"):
            query = f"{query}\n;"

        result = kql_exec(query)
        self._set_kql_option(option="auto_dataframe", value=auto_dataframe)
        if result is not None:
            if isinstance(result, pd.DataFrame):
                return result, None
            if hasattr(result, "completion_query_info") and (
                result.completion_query_info.get("StatusCode") == 0
                or result.completion_query_info.get("Text")
                == "Query completed successfully"
            ):
                data_frame = result.to_dataframe()
                if result.is_partial_table:
                    print("Warning - query returned partial results.")
                return data_frame, result

        return self._raise_query_failure(query, result)

    def _load_kql_magic(self):
        """Load KqlMagic if not loaded."""
        # KqlMagic
        print("Please wait. Loading Kqlmagic extension...", end="")
        if self._ip is not None:
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore")
                self._ip.run_line_magic("reload_ext", "Kqlmagic")
        self._loaded = True
        print("done")

    def _is_kqlmagic_loaded(self) -> bool:
        """Return true if kql magic is loaded."""
        if self._ip is not None:
            return self._ip.find_magic("kql") is not None
        return bool(kql_exec("--version"))

    @property
    def _connect_target(self) -> str:
        if self.environment == DataEnvironment.MSSentinel:
            return "Workspace"
        return "Kusto cluster"

    @staticmethod
    def _get_schema() -> Dict[str, Dict]:
        return kql_exec("--schema")

    @staticmethod
    def _get_kql_option(option):
        """Retrieve a current Kqlmagic notebook option."""
        return kql_exec(f"--config {option}").get(option)

    @staticmethod
    def _set_kql_option(option, value):
        """Set a Kqlmagic notebook option."""
        opt_val = f"'{value}'" if isinstance(value, str) else value
        return kql_exec(f"--config {option}={opt_val}")

    def _set_kql_cloud(self):
        """If cloud is set in Azure Settings override default."""
        # Check that there isn't a cloud setting in the KQLMAGIC env var
        kql_config = os.environ.get("KQLMAGIC_CONFIGURATION", "")
        if "cloud" in kql_config:
            # Set by user - we don't want to override this
            kql_cloud = self._get_kql_option("cloud")
            az_cloud = _AZ_CLOUD_MAP.get(kql_cloud, "public")
            return kql_cloud, az_cloud
        az_cloud = AzureCloudConfig().cloud
        kql_cloud = _KQL_CLOUD_MAP.get(az_cloud, "public")
        if kql_cloud != self._get_kql_option("cloud"):
            self._set_kql_option("cloud", kql_cloud)
        return kql_cloud, az_cloud

    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return date_time.isoformat(sep="T") + "Z"

    @staticmethod
    def _format_list(param_list: Iterable[Any]):
        """Return formatted list parameter."""
        fmt_list = []
        for item in param_list:
            if isinstance(item, str):
                fmt_list.append(f"'{item}'")
            else:
                fmt_list.append(f"{item}")
        return ",".join(fmt_list)

    @staticmethod
    def _raise_query_failure(query, result):
        """Raise query failure exception."""
        err_contents = []
        if hasattr(result, "completion_query_info"):
            q_info = result.completion_query_info
            if "StatusDescription" in q_info:
                err_contents = [
                    f"StatusDescription {q_info.get('StatusDescription')}",
                    f"(err_code: {result.completion_query_info.get('StatusCode')})",
                ]
            elif "Text" in q_info:
                err_contents = [f"StatusDescription {q_info.get('Text')}"]
            else:
                err_contents = [f"Unknown error type: {q_info}"]
        if not err_contents:
            err_contents = ["Unknown query error"]

        err_contents.append(f"Query:\n{query}")
        raise MsticpyDataQueryError(*err_contents)

    _WS_RGX = r"workspace\(['\"](?P<ws>[^'\"]+)"
    _TEN_RGX = r"tenant\(['\"](?P<tenant>[^'\"]+)"

    def _raise_kql_error(self, ex):
        kql_err = json.loads(ex.args[0]).get("error")
        if kql_err.get("code") == "WorkspaceNotFoundError":
            ex_mssgs = [
                "The workspace ID used to connect to Microsoft Sentinel could not be found.",
                "Please check that this is a valid workspace for your subscription",
            ]
            ws_match = re.search(self._WS_RGX, self.current_connection, re.IGNORECASE)
            if ws_match:
                ws_name = ws_match.groupdict().get("ws")
                ex_mssgs.append(f"The workspace id used was {ws_name}.")
            ex_mssgs.append(f"The full connection string was {self.current_connection}")
            raise MsticpyKqlConnectionError(*ex_mssgs, title="unknown workspace")
        raise MsticpyKqlConnectionError(
            "The service returned the following error when connecting",
            str(ex),
            title="Kql response error",
        )

    @staticmethod
    def _raise_kql_engine_error(ex):
        ex_mssgs = [
            "An error was returned from Kqlmagic KqlEngine.",
            "This can occur if you tried to connect to a second workspace using a"
            + " different tenant ID - only a single tenant ID is supported in"
            + " one notebook.",
            "Other causes of this error could be an invalid format of your"
            + " connection string",
            *(ex.args),
        ]
        raise MsticpyKqlConnectionError(*ex_mssgs, title="kql connection error")

    @staticmethod
    def _raise_adal_error(ex):
        """Adal error - usually wrong tenant ID."""
        if ex.args and ex.args[0] == "Unexpected polling state code_expired":
            raise MsticpyKqlConnectionError(
                "Authentication request was not completed.",
                title="authentication timed out",
            )

        err_response = getattr(ex, "error_response", None)
        if err_response and "error_description" in ex.error_response:
            ex_mssgs = ex.error_response["error_description"].split("\r\n")
        else:
            ex_mssgs = [f"Full error: {ex}"]
        raise MsticpyKqlConnectionError(
            *ex_mssgs, title="could not authenticate to tenant"
        )

    @staticmethod
    def _raise_authn_error(ex):
        """Raise an authentication error."""
        ex_mssgs = [
            "The authentication failed.",
            "Please check the credentials you are using and permissions on the ",
            "workspace or cluster.",
            *(ex.args),
        ]
        raise MsticpyKqlConnectionError(*ex_mssgs, title="authentication failed")

    @staticmethod
    def _raise_unknown_error(ex):
        """Raise an unknown exception."""
        raise MsticpyKqlConnectionError(
            "Another exception was returned by the service",
            *ex.args,
            f"Full exception:\n{ex}",
            title="connection failed",
        )

    def _set_az_auth_option(
        self,
        mp_az_auth: Union[bool, str, list, None],
    ):
        """
        Build connection string with auth elements.

        Parameters
        ----------
        mp_az_auth : Union[bool, str, list, None], optional
            Optional parameter directing KqlMagic to use MSTICPy Azure authentication.
            Values can be:
            - True or "default": use the settings in msticpyconfig.yaml 'Azure' section
            - auth_method: single auth method name ('msi', 'cli', 'env' or 'interactive')
            - auth_methods: list of acceptable auth methods from ('msi', 'cli',
                'env' or 'interactive')

        """
        print("Authenticating to Azure.")
        # default to default auth methods
        az_config = AzureCloudConfig()
        auth_types = az_config.auth_methods
        # override if user-supplied methods on command line
        if isinstance(mp_az_auth, str) and mp_az_auth != "default":
            auth_types = [mp_az_auth]
        elif isinstance(mp_az_auth, list):
            auth_types = mp_az_auth
        # get current credentials
        creds = az_connect(auth_methods=auth_types)
        if only_interactive_cred(creds.modern):
            print("Check your default browser for interactive sign-in prompt.")

        la_uri = _LOGANALYTICS_URL_BY_CLOUD[self.az_cloud]
        la_token_uri = f"{la_uri}.default"
        # obtain token for Log Analytics
        token = creds.modern.get_token(la_token_uri)
        # set the token values in the namespace

        la_token = {
            "access_token": token.token,
            "token_type": "Bearer",
            "resource": la_uri,
        }
        self._set_kql_option("try_token", la_token)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Threat Intel Results Browser."""
import pprint
from typing import List, Union
import pandas as pd
from IPython.core.display import HTML

from .._version import VERSION
from .nbwidgets import SelectItem

__version__ = VERSION
__author__ = "Ian Hellen"


def browse_results(
    data: pd.DataFrame, severities: Union[List[str], str, None] = None, **kwargs
) -> SelectItem:
    """
    Return TI Results list browser.

    Parameters
    ----------
    data : pd.DataFrame
        TI Results data from TIProviders
    severities : Union[List[str], str, None], optional
        A list of the severity classes to show or the string 'all'.
        By default these are ['warning', 'high'].

    Other Parameters
    ----------------
    kwargs :
        passed to SelectItem constuctor.

    Returns
    -------
    SelectItem
        SelectItem browser for TI Data.

    """
    if "height" not in kwargs:
        kwargs["height"] = "300px"

    if not isinstance(data, pd.DataFrame) or data.empty:
        raise ValueError("'data' parameter is empty or is not a dataframe.")

    opts = get_ti_select_options(ti_data=data, severities=severities)
    disp_func = ti_details_display(ti_data=data)
    if not opts and not data.empty:
        print("No results at TI Severities 'warning' or 'high'")
        print("Displaying only 'information' severity items.")
        opts = get_ti_select_options(ti_data=data, severities="all")

    return SelectItem(item_dict=opts, action=disp_func, **kwargs)


def get_ti_select_options(
    ti_data: pd.DataFrame, severities: Union[List[str], str, None] = None
):
    """Get SelectItem options for TI data."""
    ti_agg_df = _create_ti_agg_list(ti_data, severities)
    return dict(
        ti_agg_df.reset_index()
        .apply(
            lambda x: (
                f"{x.Ioc:<40} type: {x.IocType:<10}  (sev: {x.Severity})"
                + f"  providers: {x.Providers}",
                (x.Ioc, x.Providers),
            ),
            axis=1,
        )
        .values
    )


def _create_ti_agg_list(
    ti_data: pd.DataFrame, severities: Union[List[str], str, None] = None
):
    """Aggregate ti results on IoC for multiple providers."""
    if not severities:
        severities = ["warning", "high"]
    if severities == "all":
        severities = ["information", "warning", "high"]
    ti_data["Details"] = ti_data.apply(lambda x: _label_col_dict(x, "Details"), axis=1)

    return (
        ti_data[ti_data["Severity"].isin(severities)]
        .groupby(["Ioc", "IocType", "Severity"])
        .agg(
            Providers=pd.NamedAgg(
                column="Provider", aggfunc=lambda x: x.unique().tolist()
            ),
            Details=pd.NamedAgg(column="Details", aggfunc=lambda x: x.tolist()),
            Responses=pd.NamedAgg(column="RawResult", aggfunc=lambda x: x.tolist()),
            References=pd.NamedAgg(
                column="Reference", aggfunc=lambda x: x.unique().tolist()
            ),
        )
        .reset_index()
    )


def _label_col_dict(row: pd.Series, column: str):
    """Add label from the Provider column to the details."""
    if not isinstance(row[column], dict):
        return row[column]
    return (
        {row.Provider: row[column]} if row.Provider not in row[column] else row[column]
    )


def ti_details_display(ti_data):
    """Return TI Details display function."""

    def get_ti_details(ioc_prov):
        """Display TI records from individual TI entry."""
        ioc, provs = ioc_prov
        results = []
        h2_style = "border: 1px solid;background-color: DarkGray; padding: 6px"
        h3_style = "background-color: SteelBlue; padding: 6px"
        results.append(f"<h2 style='{h2_style}'>{ioc}</h2>")
        for prov in provs:
            ioc_match = ti_data[
                (ti_data["Ioc"] == ioc) & (ti_data["Provider"] == prov)
            ].iloc[0]
            results.append(
                f"<h3 style='{h3_style}'>Type: '{ioc_match.IocType}', Provider: {prov}, "
                + f"severity: {ioc_match.Severity}</h3>"
            )
            results.append("<h4>Details</h4>")
            results.append(_ti_detail_table(ioc_match.Details))
            results.append(
                f"<h4>Reference: </h4><a href='{ioc_match.Reference}'>"
                + f"{ioc_match.Reference}</a><br>"
            )

            results.append("<hr>")
            results.append("<h4>Raw Results</h4>")
            results.append(raw_results(ioc_match.RawResult))
        return HTML("".join(results))

    return get_ti_details


def raw_results(raw_result: str) -> str:
    """Create pre-formatted details for raw results."""
    fmt_details = (
        pprint.pformat(raw_result).replace("\n", "<br>").replace(" ", "&nbsp;")
    )
    return f"""
        <details>
        <summary> <u>Raw results from provider...</u></summary>
        <pre  style="font-size:11px">{fmt_details}</pre>
        </details>
        """


_TI_TABLE_STYLE = """
<style>
    .tb_ti_res {border-collapse: collapse; width: 60%; border: 1px solid #ddd !important;}
    .cell_ti {border: 1px solid #ddd !important;
        text-align: left !important; padding: 5px !important; height: 12px}
    .cell_ti_first {border: 1px solid #ddd !important; width: 25%;
        text-align: left !important; padding: 5px !important; height: 12px}
</style>
"""


def _ti_detail_table(detail_dict: dict) -> str:
    """Return table of ti details."""
    return "".join(
        [
            _TI_TABLE_STYLE,
            "<table class='tb_ti_res'>",
            *_dict_to_html(detail_dict),
            "</table>",
        ]
    )


def _dict_to_html(detail_dict):
    html_txt = []
    if not isinstance(detail_dict, dict):
        return detail_dict
    for key, val in detail_dict.items():
        html_txt.append(f"<tr class='cell_ti'><td class='cell_ti_first'>{key}</td>")
        if not isinstance(val, dict):
            html_txt.append(f"<td class='cell_ti'>{val}</td></tr>")
        else:
            html_txt.extend(_dict_to_html(val))
            html_txt.append("</tr>")
    return html_txt



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""RegistryValue Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .entity_enums import RegistryHive

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class RegistryKey(Entity):
    """
    RegistryKey Entity class.

    Attributes
    ----------
    Hive : RegistryHive
        RegistryKey Hive
    Key : str
        RegistryKey Key

    """

    ID_PROPERTIES = ["Hive", "Key"]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Hive: Optional[RegistryHive] = None
        self.Key: Optional[str] = None
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Hive.short_name if self.Hive else 'HiveNA'}\\{self.Key}"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        hive = self.Hive.short_name if self.Hive else ""
        r_key = self.Key.rsplit("\\", maxsplit=1)[-1] if self.Key else "RegKeyNA"
        return f"{hive}\\..\\{r_key}"

    _entity_schema = {
        # Hive (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.RegistryHive])
        "Hive": "RegistryHive",
        # Key (type System.String)
        "Key": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process Tree Visualization."""
from collections import namedtuple
from typing import Optional, Tuple, Union

from bokeh.io import output_notebook, show, reset_output
from bokeh.plotting import figure
from bokeh.transform import dodge, factor_cmap, linear_cmap
from bokeh.models import (
    HoverTool,
    ColumnDataSource,
    CustomJS,
    # Legend,
    # LegendItem,
    BoxSelectTool,
    RangeTool,
    ColorBar,
)

# pylint: disable=no-name-in-module
from bokeh.palettes import viridis

# pylint: enable=no-name-in-module
from bokeh.layouts import column, row
from bokeh.models import LayoutDOM
from bokeh.models.widgets import DataTable, TableColumn, DateFormatter

import numpy as np
import pandas as pd

from ..sectools.proc_tree_builder import build_process_tree, infer_schema
from ..sectools.proc_tree_schema import ProcSchema, ProcessTreeSchemaException

# pylint: disable=unused-import
from ..sectools.process_tree_utils import (  # noqa F401
    get_process_key,
    get_roots,
    get_process,
    get_parent,
    get_root,
    get_root_tree,
    get_tree_depth,
    get_children,
    get_descendents,
    get_ancestors,
    get_siblings,
    get_summary_info,
)

# pylint: enable=unused-import

from ..common.utility import check_kwargs
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

_DEFAULT_KWARGS = ["height", "title", "width", "hide_legend", "pid_fmt"]


def build_and_show_process_tree(
    data: pd.DataFrame,
    schema: ProcSchema = None,
    output_var: str = None,
    legend_col: str = None,
    **kwargs,
) -> Tuple[figure, LayoutDOM]:
    """
    Build process tree from data and plot a tree.

    Parameters
    ----------
    data : pd.DataFrame
        Window process creation or Linux Auditd events
    schema : ProcSchema
        The data schema to use for the data set, by default None
        (if None the schema is inferred)
    output_var : str, optional
        Output variable for selected items in the tree,
        by default None
    legend_col : str, optional
        The column used to color the tree items, by default None
    kwargs : Dict[str, Any]
        Additional arguments passed to plot_process_tree

    Other Parameters
    ----------------
    height : int, optional
        The height of the plot figure
        (the default is 700)
    width : int, optional
        The width of the plot figure (the default is 900)
    title : str, optional
        Title to display (the default is None)
    hide_legend : bool, optional
        Hide the legend box, even if legend_col is specified.
    pid_fmt : str, optional
        Display Process ID as 'dec' (decimal) or 'hex' (hexadecimal),
        default is 'hex'.

    Returns
    -------
    Tuple[figure, LayoutDOM]:
        figure - The main bokeh.plotting.figure
        Layout - Bokeh layout structure.

    Notes
    -----
    For full parameter set for process tree display see the
    help for plot_process_tree.

    See Also
    --------
    plot_process_tree

    """
    # Check if this table already seems to have the proc_tree metadata
    missing_cols = _check_proc_tree_schema(data)
    if missing_cols:
        data = build_process_tree(procs=data, schema=schema)

    return plot_process_tree(
        data, schema, output_var=output_var, legend_col=legend_col, **kwargs
    )


# pylint: disable=too-many-locals, too-many-statements
def plot_process_tree(  # noqa: MC0001
    data: pd.DataFrame,
    schema: ProcSchema = None,
    output_var: str = None,
    legend_col: str = None,
    show_table: bool = False,
    **kwargs,
) -> Tuple[figure, LayoutDOM]:
    """
    Plot a Process Tree Visualization.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame containing one or more Process Trees
    schema : ProcSchema, optional
        The data schema to use for the data set, by default None
        (if None the schema is inferred)
    output_var : str, optional
        Output variable for selected items in the tree,
        by default None
    legend_col : str, optional
        The column used to color the tree items, by default None
    show_table: bool
        Set to True to show a data table, by default False.

    Other Parameters
    ----------------
    height : int, optional
        The height of the plot figure
        (the default is 700)
    width : int, optional
        The width of the plot figure (the default is 900)
    title : str, optional
        Title to display (the default is None)
    hide_legend : bool, optional
        Hide the legend box, even if legend_col is specified.
    pid_fmt : str, optional
        Display Process ID as 'dec' (decimal) or 'hex' (hexadecimal),
        default is 'hex'.

    Returns
    -------
    Tuple[figure, LayoutDOM]:
        figure - The main bokeh.plotting.figure
        Layout - Bokeh layout structure.

    Raises
    ------
    ProcessTreeSchemaException
        If the data set schema is not valid for the plot.

    Notes
    -----
    The `output_var` variable will be overwritten with any selected
    values.

    """
    check_kwargs(kwargs, _DEFAULT_KWARGS)
    reset_output()
    output_notebook()

    plot_height: int = kwargs.pop("height", 700)
    plot_width: int = kwargs.pop("width", 900)
    title: str = kwargs.pop("title", "ProcessTree")
    hide_legend = kwargs.pop("hide_legend", False)
    pid_fmt = kwargs.pop("pid_fmt", "hex")

    proc_data, schema, levels, n_rows = _pre_process_tree(data, schema, pid_fmt=pid_fmt)
    if schema is None:
        raise ProcessTreeSchemaException("Could not infer data schema from data set.")

    source = ColumnDataSource(data=proc_data)
    # Get legend/color bar map
    fill_map, color_bar = _create_fill_map(source, legend_col)

    max_level = max(levels) + 3
    min_level = min(levels)

    if color_bar:
        title += " (color bar = {legend_col})"
    visible_range = int(plot_height / 35)
    y_start_range = (n_rows - visible_range, n_rows + 1)
    b_plot = figure(
        title=title,
        plot_width=plot_width,
        plot_height=plot_height,
        x_range=(min_level, max_level),
        y_range=y_start_range,
        tools=["reset", "save", "tap", "ywheel_pan"],
        toolbar_location="above",
        active_scroll="ywheel_pan",
    )

    hover = HoverTool(
        tooltips=_get_tool_tips(schema),
        formatters={f"@{schema.time_stamp}": "datetime"},
    )
    b_plot.add_tools(hover)

    # dodge to align rectangle with grid
    rect_x = dodge("Level", 1.75, range=b_plot.x_range)
    rect_plot_params = dict(
        width=3.5, height=0.95, source=source, fill_alpha=0.4, fill_color=fill_map
    )

    if color_bar:
        b_plot.add_layout(color_bar, "right")
    elif legend_col:
        rect_plot_params["legend_field"] = legend_col
    rect_plot = b_plot.rect(x=rect_x, y="Row", **rect_plot_params)
    if legend_col and not color_bar:
        b_plot.legend.title = legend_col
        b_plot.legend.label_text_font_size = "7pt"
    if hide_legend:
        b_plot.legend.visible = False

    text_props = {"source": source, "text_align": "left", "text_baseline": "middle"}

    def x_dodge(x_offset):
        return dodge("Level", x_offset, range=b_plot.x_range)

    def y_dodge(y_offset):
        return dodge("Row", y_offset, range=b_plot.y_range)

    b_plot.text(
        x=x_dodge(0.1),
        y=y_dodge(-0.2),
        text="__cmd_line$$",
        text_font_size="7pt",
        **text_props,
    )
    b_plot.text(
        x=x_dodge(0.1),
        y=y_dodge(0.25),
        text="__proc_name$$",
        text_font_size="8pt",
        **text_props,
    )
    b_plot.text(
        x=x_dodge(2.2),
        y=y_dodge(0.25),
        text="__proc_id$$",
        text_font_size="8pt",
        **text_props,
    )

    # Plot options
    _set_plot_option_defaults(b_plot)
    b_plot.xaxis.ticker = sorted(levels)
    b_plot.xgrid.ticker = sorted(levels)
    b_plot.hover.renderers = [rect_plot]  # only hover element boxes

    # Selection callback
    if output_var is not None:
        get_selected = _create_js_callback(source, output_var)
        b_plot.js_on_event("tap", get_selected)
        box_select = BoxSelectTool(callback=get_selected)
        b_plot.add_tools(box_select)

    range_tool = _create_vert_range_tool(
        data=source,
        min_y=0,
        max_y=n_rows,
        plot_range=b_plot.y_range,
        width=90,
        height=plot_height,
        x_col="Level",
        y_col="Row",
        fill_map=fill_map,
    )
    plot_elems = row(b_plot, range_tool)
    if show_table:
        data_table = _create_data_table(source, schema, legend_col)
        plot_elems = column(plot_elems, data_table)
    show(plot_elems)
    return b_plot, plot_elems


# pylint: enable=too-many-locals, too-many-statements


TreeResult = namedtuple("TreeResult", "proc_tree, schema, levels, n_rows")


def _pre_process_tree(
    proc_tree: pd.DataFrame, schema: ProcSchema = None, pid_fmt: str = "hex"
):
    """Extract dimensions and formatted values from proc_tree."""
    # Check if this table already seems to have the proc_tree metadata
    missing_cols = _check_proc_tree_schema(proc_tree)
    if missing_cols:
        proc_tree = build_process_tree(procs=proc_tree, schema=schema)

    if schema is None:
        schema = infer_schema(proc_tree)
    if schema is None:
        return TreeResult(None, None, None, None)

    _validate_plot_schema(proc_tree, schema)

    proc_tree = proc_tree.sort_values("path", ascending="True").reset_index()
    n_rows = len(proc_tree)
    proc_tree["Row"] = proc_tree.index
    proc_tree["Row"] = n_rows - proc_tree["Row"]
    proc_tree["Level"] = proc_tree["path"].str.count("/") + 1

    levels = proc_tree["Level"].unique()

    proc_tree["__proc_name$$"] = proc_tree.apply(
        lambda x: x[schema.process_name].split(schema.path_separator)[-1], axis=1
    )
    proc_tree["__proc_id$$"] = proc_tree[schema.process_id].apply(
        _pid_fmt, args=(pid_fmt,)
    )

    # trim long commandlines
    max_cmd_len = int(500 / len(levels))
    long_cmd = proc_tree[schema.cmd_line].str.len() > max_cmd_len
    proc_tree.loc[long_cmd, "__cmd_line$$"] = (
        proc_tree[long_cmd][schema.cmd_line].str[:max_cmd_len] + "..."
    )
    # replace missing cmdlines
    proc_tree.loc[~long_cmd, "__cmd_line$$"] = proc_tree[~long_cmd][
        schema.cmd_line
    ].fillna("cmdline unknown")
    return TreeResult(proc_tree=proc_tree, schema=schema, levels=levels, n_rows=n_rows)


def _pid_fmt(pid, pid_fmt):
    if pid_fmt == "hex":
        return f"PID: {pid}" if str(pid).startswith("0x") else f"PID: 0x{int(pid):x}"
    return (
        f"PID: {pid}" if not str(pid).startswith("0x") else f"PID: {int(pid, base=16)}"
    )


def _validate_plot_schema(proc_tree: pd.DataFrame, schema):
    """Validate that we have the required columns."""
    required_cols = set(
        ["path", schema.cmd_line, schema.process_name, schema.process_id]
    )
    proc_cols = set(proc_tree.columns)
    missing = required_cols - proc_cols
    if missing:
        raise ProcessTreeSchemaException(
            f"Required columns not found in data set: {','.join(missing)}"
        )


def _set_plot_option_defaults(b_plot):
    """Set default plot options."""
    b_plot.outline_line_color = None
    b_plot.grid.grid_line_color = "navy"
    b_plot.axis.axis_line_color = None
    b_plot.axis.major_tick_line_color = "navy"
    b_plot.xaxis.visible = False
    b_plot.yaxis.visible = False
    b_plot.xgrid.visible = True
    b_plot.ygrid.visible = False
    b_plot.xgrid.minor_grid_line_color = "navy"
    b_plot.xgrid.minor_grid_line_alpha = 0.1
    b_plot.xgrid.grid_line_color = "navy"
    b_plot.xgrid.grid_line_alpha = 0.1
    b_plot.axis.major_label_standoff = 0


def _get_tool_tips(schema: ProcSchema):
    """Return tool tip formatter."""
    return [
        ("Process", f"@{schema.process_name}"),
        ("PID", f"@{schema.process_id}"),
        ("CmdLine", f"@{schema.cmd_line}"),
        ("SubjUser", f"@{schema.user_name}"),
        ("SubjLgnId", f"@{schema.logon_id}"),
        ("TgtLgnId", f"@{schema.target_logon_id}"),
        ("Time", f"@{schema.time_stamp}{{%F %T.%3N}}"),
    ]


def _create_js_callback(source: ColumnDataSource, result_var: str) -> CustomJS:
    """Create and return CustomJS callback to set Python variable."""
    ret_var_js = """
        // get data source from Callback args
        var inds = source.selected.indices;
        var output = [];
        for (var i = 0; i < inds.length; i++) {
            output.push(source.data[itemkey][inds[i]])
        }
        var out_str = JSON.stringify(output);
        var py_str = `${output_var} = ${out_str}`;
        console.log(py_str);
        IPython.notebook.kernel.execute(py_str);
    """
    return CustomJS(
        args=dict(source=source, itemkey="proc_key", output_var=result_var),
        code=ret_var_js,
    )


def _create_fill_map(
    source: ColumnDataSource, source_column: str = None
) -> Tuple[Union[factor_cmap, linear_cmap], Optional[ColorBar]]:
    """Create factor map or linear map based on `source_column`."""
    fill_map = "navy"
    color_bar = None
    if source_column is None or source_column not in source.data:
        return fill_map, color_bar

    col_kind = source.data[source_column].dtype.kind
    if col_kind in ["b", "O"]:
        s_values = set(source.data[source_column])
        if np.nan in s_values:
            s_values.remove(np.nan)
        values = list(s_values)
        fill_map = factor_cmap(
            source_column, palette=viridis(max(3, len(values))), factors=values
        )
    elif col_kind in ["i", "u", "f", "M"]:
        values = [val for val in source.data[source_column] if not np.isnan(val)]
        fill_map = linear_cmap(
            field_name=source_column,
            palette=viridis(256),
            low=np.min(values),
            high=np.max(values),
        )
        color_bar = ColorBar(
            color_mapper=fill_map["transform"], width=8, location=(0, 0)  # type: ignore
        )
    return fill_map, color_bar


# pylint: disable=too-many-arguments
def _create_vert_range_tool(
    data, min_y, max_y, plot_range, width, height, x_col, y_col, fill_map="navy"
):
    """Return vertical range too for plot."""
    rng_select = figure(
        plot_width=width,
        plot_height=height,
        y_range=(min_y - 1, max_y + 1),
        toolbar_location=None,
    )

    x_dodge = dodge(x_col, -0.5)
    rng_select.rect(
        x=x_dodge,
        y=y_col,
        width=1.2,
        height=0.8,
        source=data,
        fill_alpha=0.6,
        fill_color=fill_map,
    )

    rng_select.xaxis.visible = False
    rng_select.yaxis.visible = False

    range_tool = RangeTool(y_range=plot_range)
    range_tool.overlay.fill_color = "navy"
    range_tool.overlay.fill_alpha = 0.2
    rng_select.ygrid.grid_line_color = None
    rng_select.xgrid.grid_line_color = None
    rng_select.add_tools(range_tool)
    rng_select.toolbar.active_multi = range_tool
    return rng_select


# pylint: enable=too-many-arguments


def _create_data_table(
    source: ColumnDataSource, schema: ProcSchema, legend_col: str = None
):
    """Return DataTable widget for source."""
    column_names = [
        schema.user_name,
        schema.user_id,
        schema.logon_id,
        schema.process_id,
        schema.process_name,
        schema.cmd_line,
        schema.parent_id,
        schema.parent_name,
        schema.target_logon_id,
    ]

    if legend_col and legend_col not in column_names:
        column_names.append(legend_col)

    date_fmt = "%F %T"
    columns = [
        TableColumn(
            field=schema.time_stamp,
            title=schema.time_stamp,
            formatter=DateFormatter(format=date_fmt),
        )
    ]
    columns2 = [
        TableColumn(field=col, title=col)
        for col in column_names
        if col in source.column_names
    ]

    return DataTable(source=source, columns=columns + columns2, width=950, height=150)


def _check_proc_tree_schema(data):
    """Return true if expected process tree columns are present."""
    if data.index.name != "proc_key":
        return {"proc_key"}
    expected_cols = set(
        [
            "parent_key",
            "IsRoot",
            "IsLeaf",
            "IsBranch",
            "path",
            "parent_index",
        ]
    )
    return expected_cols - set(data.columns)


# pylint: disable=too-few-public-methods
@pd.api.extensions.register_dataframe_accessor("mp_process_tree")
class ProcessTreeAccessor:
    """Pandas api extension for Process Tree."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        self._df = pandas_obj

    def plot(self, **kwargs) -> Tuple[figure, LayoutDOM]:
        """
        Build and plot a process tree.

        Parameters
        ----------
        schema : ProcSchema, optional
            The data schema to use for the data set, by default None
            (if None the schema is inferred)
        output_var : str, optional
            Output variable for selected items in the tree,
            by default None
        legend_col : str, optional
            The column used to color the tree items, by default None
        show_table: bool
            Set to True to show a data table, by default False.

        Other Parameters
        ----------------
        height : int, optional
            The height of the plot figure
            (the default is 700)
        width : int, optional
            The width of the plot figure (the default is 900)
        title : str, optional
            Title to display (the default is None)
        hide_legend : bool, optional
            Hide the legend box, even if legend_col is specified.
        pid_fmt : str, optional
            Display Process ID as 'dec' (decimal) or 'hex' (hexadecimal),
            default is 'hex'.

        Returns
        -------
        Tuple[figure, LayoutDOM]:
            figure - The main bokeh.plotting.figure
            Layout - Bokeh layout structure.

        """
        return build_and_show_process_tree(data=self._df, **kwargs)

    def build(self, schema: ProcSchema = None, **kwargs) -> pd.DataFrame:
        """
        Build process trees from the process events.

        Parameters
        ----------
        procs : pd.DataFrame
            Process events (Windows 4688 or Linux Auditd)
        schema : ProcSchema, optional
            The column schema to use, by default None
            If None, then the schema is inferred
        show_summary : bool
            Shows summary of the built tree, default is False. : bool
        debug : bool
            If True produces extra debugging output,
            by default False

        Returns
        -------
        pd.DataFrame
            Process tree dataframe.

        Notes
        -----
        It is not necessary to call this before `plot`. The process
        tree is built automatically. This is only needed if you want
        to return the processed tree data as a DataFrame

        """
        return build_process_tree(
            procs=self._df,
            schema=schema,
            show_summary=kwargs.get("show_summary", kwargs.get("show_progress", False)),
            debug=kwargs.get("debug", False),
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Other Providers Component Edit."""
from .._version import VERSION
from .ce_provider_base import CEProviders, HELP_URIS

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-ancestors, duplicate-code
class CEOtherProviders(CEProviders):
    """Other (GeoIP) providers edit component."""

    _DESCRIPTION = "Settings for Other (GeoIP) Providers"
    _COMP_PATH = "OtherProviders"
    # _HELP_TEXT inherited from base
    _HELP_URI = {
        "GeoIP Providers": (
            "https://msticpy.readthedocs.io/en/latest/"
            + "data_acquisition/GeoIPLookups.html"
        ),
        **HELP_URIS,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""TI Providers Component Edit."""
from .._version import VERSION
from .ce_provider_base import CEProviders, HELP_URIS

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-ancestors, duplicate-code
class CETIProviders(CEProviders):
    """Threat Intel providers edit component."""

    _DESCRIPTION = "Settings for Threat Intelligence Providers"
    _COMP_PATH = "TIProviders"
    # _HELP_TEXT inherited from base
    _HELP_URI = {
        "Threat Intel Providers": (
            "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
            + "TIProviders.html#configuration-file"
        ),
        **HELP_URIS,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module docstring."""
import ipywidgets as widgets

from .._version import VERSION
from ..data import DataEnvironment
from .ce_common import ITEM_LIST_LAYOUT, get_or_create_mpc_section
from .comp_edit import CEItemsBase
from .compound_ctrls import UserDefLoadComponent, UserDefQryProvCtrl
from .mp_config_control import MpConfigControls

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-ancestors
class CEAutoLoadQProvs(CEItemsBase):
    """User Default query providers edit component."""

    _COMP_PATH = "UserDefaults.QueryProviders"
    _DESCRIPTION = "User Defaults - Auto-loaded query providers"
    _HELP_TEXT = """
    Choose the query providers to load when you run nbinit.init_notebook().<br>

    There are two classes of providers - AzureSentinel workspaces and other
    providers.
    In the former case you must specify a workspace to load (by default it is
    the "Default" workspace).
    Other providers typically only have a single profile so there is no analgous
    setting<br><br>

    Query providers have two optional parameters:
    <ul>
        <li><b>alias</b> - upon loading, msticpy will create a variable based on
        this alias, prefixed with "qry_". E.g. if you set alias="local" a variable
        named "qry_local" is created in the Python/Jupyter namespace. You can use
        this variable to execute queries for this provider.
        </li>
        <li><b>connect</b> - the default behavior is to connect/authenticate
        this provider after loading it. Set this to False if you do not want
        to authenticate immediately.
        </li>
    </ul>
    """
    _HELP_URI = {
        "User Defaults": (
            "https://msticpy.readthedocs.io/en/latest/getting_started/"
            + "msticpyconfig.html#user-defaults"
        ),
        "Help on this tab": (
            "https://msticpy.readthedocs.io/en/latest/getting_started/"
            + "SettingsEditor.html#autoload-query-providers"
        ),
    }

    def __init__(self, mp_controls: MpConfigControls):
        """
        Initialize an instance of CEAutoLoad class.

        Parameters
        ----------
        mp_controls : MpConfigControls
            The config/controls/settings database

        """
        super().__init__(mp_controls)

        get_or_create_mpc_section(self.mp_controls, self._COMP_PATH)
        self.select_item.options = self._get_select_opts()
        self.select_item.layout = ITEM_LIST_LAYOUT["layout"]
        self.select_item.style = ITEM_LIST_LAYOUT["style"]
        self.select_item.description = "Providers"
        self.select_item.observe(self._select_provider, names="label")

        self.prov_options = widgets.Dropdown(
            options=self._get_available_options(),
            description="Add item",
            value=self.select_item.label,
            style=ITEM_LIST_LAYOUT["style"],
        )
        self.items_frame.children = [*(self.items_frame.children), self.prov_options]

        prov_name = self.select_item.label
        edit_ctrl = self._get_settings_ctrls(prov_name, self._COMP_PATH)
        self.edit_ctrls = (
            edit_ctrl if isinstance(edit_ctrl, widgets.Widget) else edit_ctrl.layout
        )
        self.edit_frame.children = [self.edit_ctrls]

        self.edit_buttons.btn_del.on_click(self._del_component)
        self.edit_buttons.btn_add.on_click(self._add_component)
        self.edit_buttons.btn_save.on_click(self._save_component)

    @property
    def _current_path(self):
        return f"{self._COMP_PATH}.{self.select_item.label}"

    def _select_provider(self, change):
        prov_name = change.get("new")
        edit_ctrl = self._get_settings_ctrls(prov_name, self._COMP_PATH)
        self.edit_ctrls = (
            edit_ctrl if isinstance(edit_ctrl, widgets.Widget) else edit_ctrl.layout
        )
        self.edit_frame.children = [self.edit_ctrls]
        self.mp_controls.populate_ctrl_values(f"{self._COMP_PATH}.{prov_name}")

    def _add_component(self, btn):
        del btn
        prov_name = self.prov_options.value
        if prov_name in self.select_item.options:
            self.set_status(f"This provider already exists: {prov_name}")
            return

        def_value = self._get_default_values(prov_name, self._COMP_PATH)
        if def_value is None:
            self.set_status(f"No definition for this provider was found: {prov_name}")
            return
        self._get_settings_ctrls(prov_name, self._COMP_PATH)
        self.mp_controls.save_ctrl_values(f"{self._COMP_PATH}.{prov_name}")
        self.select_item.options = sorted([prov_name, *(self.select_item.options)])
        self.select_item.label = prov_name

    def _del_component(self, btn):
        del btn
        self.mp_controls.del_value(self._current_path)
        remaining_opts = self._get_select_opts()
        self.select_item.options = remaining_opts
        if remaining_opts:
            self.select_item.label = remaining_opts[-1][0]

    def _save_component(self, btn):
        del btn
        self.mp_controls.save_ctrl_values(self._current_path)
        val_results = self.mp_controls.validate_setting(self._current_path)
        status = "  ".join(res[1] for res in val_results if not res[0])
        if status:
            self.set_status(status)

    def _get_available_options(self):
        az_sent_provs = self.mp_controls.mp_config.get("AzureSentinel")
        if az_sent_provs:
            # If we have Sentinel providers
            opt_list = [f"AzureSentinel.{ws}" for ws in az_sent_provs.get("Workspaces")]
        else:
            opt_list = []
        opt_list.extend(
            {
                DataEnvironment.parse(prov).name
                for prov in DataEnvironment.__members__
                if prov not in ("Unknown", "Kusto", "AzureSentinel", "LogAnalytics")
            }
        )
        return opt_list

    def _get_select_opts(self):
        return list(self._get_query_providers())

    def _get_query_providers(self):
        for name, settings in self.mp_controls.get_value(self._COMP_PATH).items():
            if name == "AzureSentinel":
                yield from (f"{name}.{wkspace}" for wkspace in settings)
            else:
                yield name

    def _get_settings_ctrls(self, prov_name, conf_path):
        if not prov_name:
            return widgets.VBox([], layout=self.no_border_layout("95%"))

        setting_path = f"{conf_path}.{prov_name}"
        prov_ctrl = self.mp_controls.get_control(setting_path)
        if not isinstance(prov_ctrl, UserDefQryProvCtrl):
            prov_ctrl = UserDefQryProvCtrl(prov_name)
            self.mp_controls.set_control(setting_path, prov_ctrl)

        curr_val = self.mp_controls.get_value(setting_path)
        if curr_val is None:
            curr_val = self._get_default_values(prov_name, conf_path)
        else:
            if "." in prov_name:
                prov, child = prov_name.split(".", maxsplit=1)
                curr_val = {prov: {child: curr_val}}
            else:
                curr_val = {prov_name: curr_val}

        prov_ctrl.value = curr_val

        return prov_ctrl

    def _get_default_values(self, prov_name, conf_path):
        wkspace = None
        if "." in prov_name:
            defn_name, wkspace = prov_name.split(".", maxsplit=1)
        else:
            defn_name = "provider"
        defn_path = f"{conf_path}.{defn_name}"
        prov_defn = self.mp_controls.get_defn(defn_path)
        if prov_defn is None:
            self.set_status(f"No definition for this provider was found: {prov_name}")
            return {}
        def_settings = {}
        for setting, defn in prov_defn.items():
            st_type, st_opts = defn if isinstance(defn, tuple) else "str", {}
            if st_type == "bool":
                def_settings[setting] = st_opts.get("default", False)
            elif st_type == "str":
                def_settings[setting] = st_opts.get("default")
        if wkspace:
            return {defn_name: {wkspace: def_settings}}
        return {defn_name: def_settings}


# pylint: disable=too-many-ancestors
class CEAutoLoadComps(CEAutoLoadQProvs):
    """User Default load components edit component."""

    _COMP_PATH = "UserDefaults.LoadComponents"
    _DESCRIPTION = "User Defaults - Auto-loaded components"
    _HELP_TEXT = """
    Choose the other components to load when you run nbinit.init_notebook().<br>

    This is related to the QueryProvider auto-load section but the parameters
    here are different.<br>
    Available components include the following:<br>
    (the names in parantheses are the names of the global variables used
    when creating these components - use theses to perform any operations
    required on the component.)
    <ul>
        <li>TILookup (ti_lookup) - Threat Intelligence lookups</li>
        <li>GeoIpLookup (geoip) - Geolocation of IP address lookups
        Note: you must specify which GeoIP Provider you wish to use.
        </li>
        <li>Notebooklets (nb) - Load the MSTIC notebooklets package. This must be
        installed in your notebook environment. You must specificy a default
        query provider for this component.
        </li>
        <li>Pivot (pivot) - Load Pivot functions into entities.</li>
        <li>AzureData (az_data) - Load the AzureData component. Optionally specify
        the authentication options you want to use.</li>
        <li>AzureSentinelAPI (azs_api) - Load the AzureSentinel API component. Optionally
        specify the authentication options you want to use.</li>
    </ul>

    The last two components also support a "connect" parameter. If this is not
    set the default is to initiate authentication to the service as soon as
    the component is loaded. set connect=False to prevent this.

    """
    _HELP_URI = {
        "User Defaults": (
            "https://msticpy.readthedocs.io/en/latest/getting_started/"
            + "msticpyconfig.html#user-defaults"
        ),
        "Help on this tab": (
            "https://msticpy.readthedocs.io/en/latest/getting_started/"
            + "SettingsEditor.html#autoload-component"
        ),
    }

    def _get_available_options(self):
        return list(self.mp_controls.get_defn(self._COMP_PATH).keys())

    def _get_select_opts(self):
        return list(self.mp_controls.get_value(self._COMP_PATH).keys())

    def _get_settings_ctrls(self, prov_name, conf_path):
        if not prov_name:
            return widgets.VBox([], layout=self.no_border_layout("95%"))

        setting_path = f"{conf_path}.{prov_name}"
        prov_ctrl = self.mp_controls.get_control(setting_path)
        if not isinstance(prov_ctrl, UserDefLoadComponent):
            prov_ctrl = UserDefLoadComponent(
                self.mp_controls, prov_name, self._COMP_PATH
            )
            self.mp_controls.set_control(setting_path, prov_ctrl)

        curr_val = self.mp_controls.get_value(setting_path)
        prov_ctrl.value = curr_val
        return prov_ctrl

    def _get_default_values(self, prov_name, conf_path):
        defn_path = f"{conf_path}.{prov_name}"
        prov_defn = self.mp_controls.get_defn(defn_path)
        def_settings = {}
        if prov_defn is None:
            return {"alias": None}
        for setting, defn in prov_defn.items():
            st_type, st_opts = defn if isinstance(defn, tuple) else "str", {}
            if st_type == "str":
                def_settings[setting] = st_opts.get("default")
            elif st_type == "bool":
                def_settings[setting] = st_opts.get("default", False)
        return def_settings



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""IpAddress Entity class."""
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Any, List, Mapping, Optional, Union

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .geo_location import GeoLocation
from .threat_intelligence import Threatintelligence

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name, too-many-instance-attributes


@export
class IpAddress(Entity):
    """
    IPAddress Entity class.

    Attributes
    ----------
    Address : str
        IpAddress Address
    Location : GeoLocation
        IpAddress Location
    ThreatIntelligence : List[Threatintelligence]
        IpAddress ThreatIntelligence

    """

    ID_PROPERTIES = ["Address"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Address: str = ""
        self.Location: Optional[GeoLocation] = None
        self.ThreatIntelligence: List[Threatintelligence] = []
        super().__init__(src_entity=src_entity, **kwargs)

        if src_event is not None and "Location" in src_event:
            self.Location = GeoLocation(src_event["Location"])
        if src_event is not None:
            if "IpAddress" in src_event:
                self.Address = src_event["IpAddress"]
            elif "Address" in src_event:
                self.Address = src_event["Address"]

    @property
    def ip_address(self) -> Union[IPv4Address, IPv6Address, None]:
        """Return a python IP address object from the entity property."""
        try:
            return ip_address(self.Address)
        except ValueError:
            return None

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return (
            f"{self.Address} - {self.Location.CountryCode}"
            if self.Location
            else self.Address
        )

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.Address or self.__class__.__name__

    _entity_schema = {
        # Address (type System.String)
        "Address": None,
        # Location (type Microsoft.Azure.Security.Detection.AlertContracts
        # .V3.ContextObjects.GeoLocation)
        "Location": "GeoLocation",
        # ThreatIntelligence (type System.Collections.Generic.List`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3
        # .ContextObjects.ThreatIntelligence])
        "ThreatIntelligence": (list, "Threatintelligence"),
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }


# Alias for IpAddress
Ip = IpAddress



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""NetworkConnection Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .ip_address import IpAddress

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class NetworkConnection(Entity):
    """
    NetworkConnection Entity class.

    Attributes
    ----------
    SourceAddress : IPAddress
        NetworkConnection SourceAddress
    SourcePort : int
        NetworkConnection SourcePort
    DestinationAddress : IPAddress
        NetworkConnection DestinationAddress
    DestinationPort : int
        NetworkConnection DestinationPort
    Protocol : str
        NetworkConnection Protocol


    """

    ID_PROPERTIES = [
        "SourceAddress",
        "SourcePort",
        "DestinationAddress",
        "DestinationPort",
    ]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.SourceAddress: Optional[IpAddress] = None
        self.SourcePort: Optional[int] = None
        self.DestinationAddress: Optional[IpAddress] = None
        self.DestinationPort: Optional[int] = None
        self.Protocol: Optional[str] = None
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return (
            f"{self.SourceAddress}:{self.SourcePort} [{self.Protocol}]-> "
            "{self.DestinationAddress}:{self.DestinationPort}"
        )

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.__class__.__name__

    _entity_schema = {
        # SourceAddress (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "SourceAddress": "IpAddress",
        # SourcePort (type System.Nullable`1[System.Int32])
        "SourcePort": None,
        # DestinationAddress (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "DestinationAddress": "IpAddress",
        # DestinationPort (type System.Nullable`1[System.Int32])
        "DestinationPort": None,
        # Protocol (type System.Nullable`1[System.Net.Sockets.ProtocolType])
        "Protocol": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
VirusTotal Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import datetime as dt
from typing import Any, Tuple, Dict

from .ti_provider_base import LookupResult, TISeverity
from .http_base import HttpProvider, IoCLookupParams
from ...common.utility import export
from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_DEF_HEADERS = {"User-Agent": "VirusTotal", "Content-Type": "application/json"}
_GZIP_HEADERS = {"Accept-Encoding": "gzip, deflate"}


@export
class VirusTotal(HttpProvider):
    """VirusTotal Lookup."""

    _BASE_URL = "https://www.virustotal.com/"

    _PARAMS = {"apikey": "{API_KEY}"}
    _IOC_QUERIES = {
        "ipv4": IoCLookupParams(
            path="vtapi/v2/ip-address/report",
            params={**_PARAMS, "ip": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "dns": IoCLookupParams(
            path="vtapi/v2/domain/report",
            params={**_PARAMS, "domain": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "file_hash": IoCLookupParams(
            path="vtapi/v2/file/report",
            params={**_PARAMS, "resource": "{observable}"},
            headers={**_DEF_HEADERS, **_GZIP_HEADERS},
        ),
        "url": IoCLookupParams(
            path="vtapi/v2/url/report",
            params={**_PARAMS, "resource": "{observable}"},
            headers={**_DEF_HEADERS, **_GZIP_HEADERS},
        ),
    }

    # pylint: disable=duplicate-code, too-many-branches
    # aliases
    _IOC_QUERIES["md5_hash"] = _IOC_QUERIES["file_hash"]
    _IOC_QUERIES["sha1_hash"] = _IOC_QUERIES["file_hash"]
    _IOC_QUERIES["sha256_hash"] = _IOC_QUERIES["file_hash"]

    _REQUIRED_PARAMS = ["API_KEY"]

    _VT_DETECT_RESULTS = {
        "detected_urls": ("url", "scan_date"),
        "detected_downloaded_samples": ("sha256", "date"),
        "detected_communicating_samples": ("sha256", "date"),
    }

    def parse_results(self, response: LookupResult) -> Tuple[bool, TISeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        Tuple[bool, TISeverity, Any]
            bool = positive or negative hit
            TISeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(response.raw_result, dict):
            return False, TISeverity.information, "Not found."

        result_dict = {
            "verbose_msg": response.raw_result.get("verbose_msg", None),
            "response_code": response.raw_result.get("response_code", None),
            "positives": 0,
        }

        if response.ioc_type in [
            "url",
            "md5_hash",
            "sha1_hash",
            "sha256_hash",
            "file_hash",
        ]:
            result_dict["resource"] = response.raw_result.get("resource", None)
            result_dict["permalink"] = response.raw_result.get("permalink", None)
            result_dict["positives"] = response.raw_result.get("positives", 0)

        else:
            for hit_type, params in self._VT_DETECT_RESULTS.items():
                if hit_type in response.raw_result:
                    self._extract_url_results(
                        response=response,
                        result_dict=result_dict,
                        hit_type=hit_type,
                        item_type=params[0],
                        date_name=params[1],
                    )

        if "positives" in result_dict:
            if result_dict["positives"] > 1:
                severity = TISeverity.high
            elif result_dict["positives"] > 0:
                severity = TISeverity.warning
            else:
                severity = TISeverity.information
        else:
            severity = TISeverity.unknown

        return True, severity, result_dict

    # pylint: enable=duplicate-code, too-many-branches

    @staticmethod
    def _extract_url_results(
        response: LookupResult,
        result_dict: Dict[str, Any],
        hit_type: str,
        item_type: str,
        date_name: str,
    ):
        if not isinstance(response.raw_result, dict):
            return
        time_scope = dt.datetime.now() - dt.timedelta(days=30)
        result_dict[hit_type] = [
            item[item_type]
            for item in response.raw_result[hit_type]
            if item_type in item
            and dt.datetime.strptime(item[date_name], "%Y-%m-%d %H:%M:%S") > time_scope
        ]
        # positives are listed per detected_url so we need to
        # pull those our and sum them.
        positives = sum(
            item["positives"]
            for item in response.raw_result[hit_type]
            if "positives" in item
            and dt.datetime.strptime(item[date_name], "%Y-%m-%d %H:%M:%S") > time_scope
        )
        result_dict["positives"] += positives
